@@ -1,9 +1,9 @@
 import os.path
-import sys
 import platform
-from typing import TYPE_CHECKING, cast, Optional
-import traceback
 import logging
+import sys
+import traceback
+from typing import TYPE_CHECKING, Optional, cast
 
 SINGLE_INSTANCE_DEFAULT = True
 
@@ -43,6 +43,23 @@ def _compute_thonny_user_dir():
         return os.path.join(data_home, "Thonny")
 
 
+def _read_configured_debug_mode():
+    if not os.path.exists(CONFIGURATION_FILE):
+        return False
+
+    try:
+        with open(CONFIGURATION_FILE, encoding="utf-8") as fp:
+            for line in fp:
+                if "debug_mode" in line and "True" in line:
+                    return True
+        return False
+    except Exception:
+        import traceback
+
+        traceback.print_exc()
+        return False
+
+
 def is_portable():
     # it can be explicitly declared as portable or shared ...
     portable_marker_path = os.path.join(os.path.dirname(sys.executable), "portable_thonny.ini")
@@ -82,11 +99,18 @@ def get_version():
 
 THONNY_USER_DIR = _compute_thonny_user_dir()
 CONFIGURATION_FILE = os.path.join(THONNY_USER_DIR, "configuration.ini")
+_CONFIGURED_DEBUG = _read_configured_debug_mode()
 
 
-def _get_ipc_file_path():
+_IPC_FILE = None
+
+
+def get_ipc_file_path():
+    global _IPC_FILE
+    if _IPC_FILE:
+        return _IPC_FILE
+
     from thonny import misc_utils
-    import getpass
 
     if platform.system() == "Windows":
         base_dir = misc_utils.get_local_appdata_dir()
@@ -98,24 +122,29 @@ def _get_ipc_file_path():
     if not base_dir or not os.path.exists(base_dir):
         base_dir = THONNY_USER_DIR
 
-    ipc_dir = os.path.join(base_dir, "thonny-%s" % getpass.getuser())
+    for name in ("LOGNAME", "USER", "LNAME", "USERNAME"):
+        if name in os.environ:
+            username = os.environ.get(name)
+            break
+    else:
+        username = os.path.basename(os.path.expanduser("~"))
+
+    ipc_dir = os.path.join(base_dir, "thonny-%s" % username)
     os.makedirs(ipc_dir, exist_ok=True)
 
     if not platform.system() == "Windows":
         os.chmod(ipc_dir, 0o700)
 
-    return os.path.join(ipc_dir, "ipc.sock")
-
-
-IPC_FILE = _get_ipc_file_path()
+    _IPC_FILE = os.path.join(ipc_dir, "ipc.sock")
+    return _IPC_FILE
 
 
 def _check_welcome():
     from thonny import misc_utils
 
     if not os.path.exists(CONFIGURATION_FILE) and not misc_utils.running_on_rpi():
-        from thonny.first_run import FirstRunWindow
         from thonny.config import ConfigurationManager
+        from thonny.first_run import FirstRunWindow
 
         mgr = ConfigurationManager(CONFIGURATION_FILE)
 
@@ -127,8 +156,8 @@ def _check_welcome():
 
 
 def launch():
-    import gettext
     import runpy
+    import socket
 
     if sys.executable.endswith("thonny.exe"):
         # otherwise some library may try to run its subprocess with thonny.exe
@@ -143,19 +172,19 @@ def launch():
     except ImportError:
         pass
 
-    gettext.install("thonny", "locale")
-    _prepare_thonny_user_dir()
+    prepare_thonny_user_dir()
+    _configure_frontend_logging()
 
     if not _check_welcome():
-        return
+        return 0
 
     if _should_delegate():
         try:
             _delegate_to_existing_instance(sys.argv[1:])
             print("Delegated to an existing Thonny instance. Exiting now.")
             return 0
-        except Exception:
-            traceback.print_exc()
+        except (ConnectionRefusedError, socket.timeout):
+            pass
 
     # Did not or could not delegate
 
@@ -170,9 +199,9 @@ def launch():
         return 0
 
     except SystemExit as e:
-        from tkinter import messagebox
+        from tkinter import messagebox, _default_root
 
-        messagebox.showerror("System exit", str(e))
+        messagebox.showerror("System exit", str(e), master=_default_root)
         return -1
 
     except Exception:
@@ -192,7 +221,7 @@ def launch():
     return 0
 
 
-def _prepare_thonny_user_dir():
+def prepare_thonny_user_dir():
     if not os.path.exists(THONNY_USER_DIR):
         os.makedirs(THONNY_USER_DIR, mode=0o700, exist_ok=True)
 
@@ -219,9 +248,9 @@ def _prepare_thonny_user_dir():
 
 
 def _should_delegate():
-    if not os.path.exists(IPC_FILE):
+    if not os.path.exists(get_ipc_file_path()):
         # no previous instance
-        return
+        return False
 
     from thonny.config import try_load_configuration
 
@@ -231,9 +260,10 @@ def _should_delegate():
 
 
 def _delegate_to_existing_instance(args):
-    from thonny import workbench
     import socket
 
+    from thonny import workbench
+
     transformed_args = []
     for arg in args:
         if not arg.startswith("-"):
@@ -243,11 +273,9 @@ def _delegate_to_existing_instance(args):
 
     try:
         sock, secret = _create_client_socket()
-    except Exception:
+    except (ConnectionRefusedError, socket.timeout):
         # Maybe the lock is abandoned
-        print("Trying to remove lock")
-        os.remove(IPC_FILE)
-        print("Successfully removed abandoned lock")
+        os.remove(get_ipc_file_path())
         raise
 
     data = repr((secret, transformed_args)).encode(encoding="utf_8")
@@ -273,7 +301,7 @@ def _create_client_socket():
     timeout = 2.0
 
     if platform.system() == "Windows":
-        with open(IPC_FILE, "r") as fp:
+        with open(get_ipc_file_path(), "r") as fp:
             port = int(fp.readline().strip())
             secret = fp.readline().strip()
 
@@ -282,12 +310,48 @@ def _create_client_socket():
     else:
         client_socket = socket.socket(socket.AF_UNIX)  # @UndefinedVariable
         client_socket.settimeout(timeout)
-        client_socket.connect(IPC_FILE)
+        client_socket.connect(get_ipc_file_path())
         secret = ""
 
     return client_socket, secret
 
 
+def _configure_frontend_logging() -> None:
+    _configure_logging("frontend.log")
+
+
+def configure_backend_logging() -> None:
+    _configure_logging("backend.log", logging.WARNING)
+
+
+def _configure_logging(filename, console_level=None):
+    logFormatter = logging.Formatter("%(levelname)-7s %(name)s: %(message)s")
+
+    # NB! Don't mess with the root logger, because (CPython) backend runs user code
+    thonny_root_logger = logging.getLogger("thonny")
+    thonny_root_logger.setLevel(_choose_logging_level())
+    thonny_root_logger.propagate = False  # otherwise it will be also reported by IDE-s root logger
+
+    log_file = os.path.join(THONNY_USER_DIR, filename)
+    file_handler = logging.FileHandler(log_file, encoding="UTF-8", mode="w")
+    file_handler.setFormatter(logFormatter)
+    thonny_root_logger.addHandler(file_handler)
+
+    console_handler = logging.StreamHandler(sys.stdout)
+    console_handler.setFormatter(logFormatter)
+    if console_level is None:
+        console_level = _choose_logging_level()
+    console_handler.setLevel(console_level)
+    thonny_root_logger.addHandler(console_handler)
+
+    thonny_root_logger.info("Thonny version: %s", get_version())
+
+    import faulthandler
+
+    fault_out = open(os.path.join(THONNY_USER_DIR, "frontend_faults.log"), mode="w")
+    faulthandler.enable(fault_out)
+
+
 def set_dpi_aware():
     # https://stackoverflow.com/questions/36134072/setprocessdpiaware-seems-not-to-work-under-windows-10
     # https://bugs.python.org/issue33656
@@ -307,9 +371,9 @@ if TYPE_CHECKING:
     # Following imports are required for MyPy
     # http://mypy.readthedocs.io/en/stable/common_issues.html#import-cycles
     import thonny.workbench
-    from thonny.workbench import Workbench
     from thonny.running import Runner
     from thonny.shell import ShellView
+    from thonny.workbench import Workbench
 
 _workbench = None  # type: Optional[Workbench]
 
@@ -321,6 +385,35 @@ def get_workbench() -> "Workbench":
 _runner = None  # type: Optional[Runner]
 
 
+def set_logging_level(level=None):
+    if level is None:
+        level = _choose_logging_level()
+
+    logging.getLogger("thonny").setLevel(level)
+
+
+def _choose_logging_level():
+    if in_debug_mode():
+        return logging.DEBUG
+    else:
+        return logging.INFO
+
+
+def in_debug_mode() -> bool:
+    # Value may be something other than string when it is set in Python code
+    return (
+        os.environ.get("THONNY_DEBUG", False)
+        in [
+            "1",
+            1,
+            "True",
+            True,
+            "true",
+        ]
+        or _CONFIGURED_DEBUG
+    )
+
+
 def get_runner() -> "Runner":
     return cast("Runner", _runner)
 
