@@ -1,4 +1,5 @@
 import datetime
+import logging
 import os.path
 import subprocess
 import time
@@ -26,6 +27,8 @@ ROOT_NODE_ID = ""
 
 HIDDEN_FILES_OPTION = "file.show_hidden_files"
 
+logger = logging.getLogger(__name__)
+
 
 class BaseFileBrowser(ttk.Frame):
     def __init__(self, master, show_expand_buttons=True):
@@ -827,7 +830,7 @@ class BaseFileBrowser(ttk.Frame):
             parent = self.current_focus
         else:
             if self.get_selected_kind() == "file":
-                # dirname does the right thing even if parent is Linux path and runnning on Windows
+                # dirname does the right thing even if parent is Linux path and running on Windows
                 parent = os.path.dirname(parent)
 
         name = ask_string(
@@ -897,7 +900,15 @@ class BaseLocalFileBrowser(BaseFileBrowser):
         get_workbench().get_editor_notebook().show_file(path)
 
     def open_path_with_system_app(self, path):
-        open_with_default_app(path)
+        try:
+            open_with_default_app(path)
+        except Exception as e:
+            logger.error("Could not open %r in system app", path, exc_info=e)
+            messagebox.showerror(
+                "Error",
+                "Could not open '%s' in system app\nError: %s" % (path, e),
+                parent=self.winfo_toplevel(),
+            )
 
     def on_window_focus_in(self, event=None):
         self.refresh_tree()
@@ -906,7 +917,6 @@ class BaseLocalFileBrowser(BaseFileBrowser):
         if event["operation"] in ["save", "delete"]:
             self.refresh_tree()
             self.select_path_if_visible(event["path"])
-            # TODO: select this file in tree?
 
     def request_fs_info(self, path):
         if path == "":
