@@ -2,17 +2,24 @@
 
 import os.path
 import platform
+import queue
 import shlex
-import shutil
 import subprocess
 import sys
+import threading
 import time
 from typing import Optional, Sequence, Tuple
 
+PASSWORD_METHOD = "password"
+PUBLIC_KEY_NO_PASS_METHOD = "public-key (without passphrase)"
+PUBLIC_KEY_WITH_PASS_METHOD = "public-key (with passphrase)"
+
 
 def delete_dir_try_hard(path: str, hardness: int = 5) -> None:
     # Deleting the folder on Windows is not so easy task
     # http://bugs.python.org/issue15496
+    import shutil
+
     for i in range(hardness):
         if os.path.exists(path):
             time.sleep(i * 0.5)
@@ -46,14 +53,7 @@ def running_on_rpi() -> bool:
 
 def list_volumes(skip_letters=set()) -> Sequence[str]:
     "Adapted from https://github.com/ntoll/uflash/blob/master/uflash.py"
-    if os.name == "posix":
-        # 'posix' means we're on Linux or OSX (Mac).
-        # Call the unix "mount" command to list the mounted volumes.
-        mount_output = subprocess.check_output("mount").splitlines()
-        return [x.split()[2].decode("utf-8") for x in mount_output]
-
-    elif os.name == "nt":
-        # 'nt' means we're on Windows.
+    if sys.platform == "win32":
         import ctypes
 
         #
@@ -76,8 +76,10 @@ def list_volumes(skip_letters=set()) -> Sequence[str]:
         finally:
             ctypes.windll.kernel32.SetErrorMode(old_mode)  # @UndefinedVariable
     else:
-        # No support for unknown operating systems.
-        raise NotImplementedError('OS "{}" not supported.'.format(os.name))
+        # 'posix' means we're on Linux or OSX (Mac).
+        # Call the unix "mount" command to list the mounted volumes.
+        mount_output = subprocess.check_output("mount").splitlines()
+        return [x.split()[2].decode("utf-8") for x in mount_output]
 
 
 def get_win_volume_name(path: str) -> str:
@@ -87,14 +89,24 @@ def get_win_volume_name(path: str) -> str:
     the given disk/device.
     Code from http://stackoverflow.com/a/12056414
     """
-    import ctypes
+    if sys.platform == "win32":
+        import ctypes
 
-    vol_name_buf = ctypes.create_unicode_buffer(1024)
-    ctypes.windll.kernel32.GetVolumeInformationW(  # @UndefinedVariable
-        ctypes.c_wchar_p(path), vol_name_buf, ctypes.sizeof(vol_name_buf), None, None, None, None, 0
-    )
-    assert isinstance(vol_name_buf.value, str)
-    return vol_name_buf.value
+        vol_name_buf = ctypes.create_unicode_buffer(1024)
+        ctypes.windll.kernel32.GetVolumeInformationW(  # @UndefinedVariable
+            ctypes.c_wchar_p(path),
+            vol_name_buf,
+            ctypes.sizeof(vol_name_buf),
+            None,
+            None,
+            None,
+            None,
+            0,
+        )
+        assert isinstance(vol_name_buf.value, str)
+        return vol_name_buf.value
+    else:
+        raise RuntimeError("Only meant for Windows")
 
 
 def find_volumes_by_name(volume_name: str, skip_letters={"A"}) -> Sequence[str]:
@@ -110,15 +122,19 @@ def find_volumes_by_name(volume_name: str, skip_letters={"A"}) -> Sequence[str]:
 
 
 def find_volume_by_name(
-    volume_name: str, not_found_msg: Optional[str] = None, found_several_msg: Optional[str] = None
+    volume_name: str,
+    not_found_msg: Optional[str] = None,
+    found_several_msg: Optional[str] = None,
+    parent=None,
 ) -> Optional[str]:
+    from thonny.languages import tr
 
     # Can't translate in the header as _ may not be available at import time
     if not_found_msg is None:
-        not_found_msg = _("Could not find disk '%s'. Do you want to locate it yourself?")
+        not_found_msg = tr("Could not find disk '%s'. Do you want to locate it yourself?")
 
     if found_several_msg is None:
-        found_several_msg = _("Found several '%s' disks. Do you want to choose one yourself?")
+        found_several_msg = tr("Found several '%s' disks. Do you want to choose one yourself?")
 
     volumes = find_volumes_by_name(volume_name)
     if len(volumes) == 1:
@@ -129,13 +145,13 @@ def find_volume_by_name(
         else:
             msg = found_several_msg % volume_name
 
+        import tkinter as tk
         from tkinter.messagebox import askyesno
+
         from thonny.ui_utils import askdirectory
 
-        import tkinter as tk
-
-        if askyesno(_("Can't find suitable disk"), msg):
-            path = askdirectory(master=tk._default_root)
+        if askyesno(tr("Can't find suitable disk"), msg, master=parent):
+            path = askdirectory(parent=parent)
             if path:
                 return path
 
@@ -149,19 +165,6 @@ def shorten_repr(original_repr: str, max_len: int = 1000) -> str:
         return original_repr
 
 
-def get_python_version_string(version_info: Optional[Tuple] = None) -> str:
-    if version_info is None:
-        version_info = sys.version_info
-
-    result = ".".join(map(str, version_info[:3]))
-    if version_info[3] != "final":
-        result += "-" + version_info[3]
-
-    result += " (" + ("64" if sys.maxsize > 2 ** 32 else "32") + " bit)\n"
-
-    return result
-
-
 def _win_get_used_memory():
     # http://code.activestate.com/recipes/578513-get-memory-usage-of-windows-processes-using-getpro/
     import ctypes
@@ -219,7 +222,7 @@ def _unix_get_used_memory():
     "TODO:"
 
 
-def construct_cmd_line(parts, safe_tokens=[]):
+def construct_cmd_line(parts, safe_tokens=[]) -> str:
     def quote(s):
         if s in safe_tokens:
             return s
@@ -229,6 +232,21 @@ def construct_cmd_line(parts, safe_tokens=[]):
     return " ".join(map(quote, parts))
 
 
+def user_friendly_python_command_line(cmd):
+    if "-m" in cmd:
+        cmd = cmd[cmd.index("-m") + 1 :]
+
+    lines = [""]
+    for item in cmd:
+        if lines[-1] and len(lines[-1] + " " + item) > 60:
+            lines.append("")
+        lines[-1] = (lines[-1] + " " + item).strip()
+
+    return "\n".join(lines)
+
+    return subprocess.list2cmdline(cmd)
+
+
 def parse_cmd_line(s):
     return shlex.split(s, posix=True)
 
@@ -236,7 +254,7 @@ def parse_cmd_line(s):
 def levenshtein_distance(s1, s2):
     # https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance#Python
     if len(s1) < len(s2):
-        return levenshtein_distance(s2, s1)
+        return levenshtein_distance(s2, s1)  # pylint: disable=arguments-out-of-order
 
     # len(s1) >= len(s2)
     if len(s2) == 0:
@@ -285,15 +303,15 @@ def levenshtein_damerau_distance(s1, s2, maxDistance):
     #  match a, change e->i change i->e => aie
     #  Damarau-Levenshtein has a cost of 1
     #  match a, transpose ei to ie => aie
-    transpositionRow = None
-    prevRow = None
+    transpositionRow = []
+    prevRow = []
 
     #  build first leven matrix row
     #  The first row represents transformation from an empty string
     #  to the shorter string making it static [0-n]
     #  since this row is static we can set it as
     #  curRow and start computation at the second row or index 1
-    curRow = [x for x in range(0, l1 + 1)]
+    curRow = list(range(0, l1 + 1))
 
     # use second length to loop through all the rows being built
     # we start at row one
@@ -383,7 +401,7 @@ class TimeHelper:
 
 def copy_to_clipboard(data):
     if running_on_windows():
-        return _copy_to_windows_clipboard(data)
+        _copy_to_windows_clipboard(data)
     elif running_on_mac_os():
         command = ["pbcopy"]
     else:
@@ -472,3 +490,25 @@ def get_roaming_appdata_dir():
 
 def get_local_appdata_dir():
     return _get_known_folder(28)
+
+
+class PopenWithOutputQueues(subprocess.Popen):
+    def __init__(self, *args, **kw):
+        super().__init__(*args, **kw)
+        self.stdout_queue = queue.Queue()
+        self.stderr_queue = queue.Queue()
+
+        for stream, target_queue in [
+            (self.stdout, self.stdout_queue),
+            (self.stderr, self.stderr_queue),
+        ]:
+            threading.Thread(
+                target=self._listen_thread, args=[stream, target_queue], daemon=True
+            ).start()
+
+    def _listen_thread(self, stream, target_queue: queue.Queue):
+        while True:
+            data = stream.readline()
+            if data == "":
+                break
+            target_queue.put(data)
