@@ -29,6 +29,7 @@ def default_light() -> SyntaxThemeSettings:
             "background": open_string_bg,
             "font": "EditorFont",
         },
+        "tab": {"background": "#f5ecd7"},
         "keyword": {"foreground": "#7f0055", "font": "BoldEditorFont"},
         "builtin": {"foreground": "#7f0055"},
         "number": {"foreground": "#B04600"},
@@ -153,6 +154,7 @@ def default_dark() -> SyntaxThemeSettings:
             "background": open_string_bg,
             "font": "EditorFont",
         },
+        "tab": {"background": "#424034"},
         "builtin": {"foreground": "#A9B1C9"},
         "keyword": {"foreground": "#A9B1C9", "font": "BoldEditorFont"},
         "number": {"foreground": "#FFCABF"},
