@@ -11,7 +11,8 @@ class LocalsHighlighter:
         self._update_scheduled = False
 
     def get_positions(self):
-        tree = jedi_utils.import_python_tree()
+        from jedi import parser_utils
+        from parso.python import tree
 
         locs = []
 
@@ -60,7 +61,7 @@ class LocalsHighlighter:
         source = self.text.get("1.0", "end")
         module = jedi_utils.parse_source(source)
         for child in module.children:
-            if isinstance(child, tree.BaseNode) and jedi_utils.is_scope(child):
+            if isinstance(child, tree.BaseNode) and parser_utils.is_scope(child):
                 process_scope(child)
 
         loc_pos = set(
@@ -92,7 +93,7 @@ class LocalsHighlighter:
     def update(self):
         self.text.tag_remove("local_name", "1.0", "end")
 
-        if get_workbench().get_option("view.locals_highlighting"):
+        if get_workbench().get_option("view.locals_highlighting") and self.text.is_python_text():
             try:
                 highlight_positions = self.get_positions()
                 self._highlight(highlight_positions)
@@ -101,6 +102,10 @@ class LocalsHighlighter:
 
 
 def update_highlighting(event):
+    if not get_workbench().ready:
+        # don't slow down loading process
+        return
+
     assert isinstance(event.widget, tk.Text)
     text = event.widget
 
@@ -111,10 +116,6 @@ def update_highlighting(event):
 
 
 def load_plugin() -> None:
-    if jedi_utils.get_version_tuple() < (0, 9):
-        logging.warning("Jedi version is too old. Disabling locals marker")
-        return
-
     wb = get_workbench()
     wb.set_default("view.locals_highlighting", False)
     wb.bind_class("CodeViewText", "<<TextChange>>", update_highlighting, True)
