@@ -12,6 +12,7 @@ from thonny.common import (
     CommandToBackend,
     EOFCommand,
     ImmediateCommand,
+    InlineCommand,
 )
 from thonny.languages import tr
 from thonny.plugins.backend_config_page import (
@@ -25,6 +26,8 @@ from thonny.ui_utils import (
     create_url_label,
 )
 
+logger = logging.getLogger(__name__)
+
 DEFAULT_WEBREPL_URL = "ws://192.168.4.1:8266/"
 
 VIDS_PIDS_TO_AVOID_IN_GENERIC_BACKEND = set()
@@ -70,6 +73,7 @@ class MicroPythonProxy(SubprocessProxy):
             "validate_time": get_workbench().get_option(
                 self.backend_name + ".validate_time", False
             ),
+            "utc_clock": get_workbench().get_option(self.backend_name + ".utc_clock", False),
         }
         return result
 
@@ -144,8 +148,10 @@ class BareMetalMicroPythonProxy(MicroPythonProxy):
             "port": self._port,
             "dtr": get_workbench().get_option(self.backend_name + ".dtr"),
             "rts": get_workbench().get_option(self.backend_name + ".rts"),
+            "submit_mode": get_workbench().get_option(self.backend_name + ".submit_mode"),
             "api_stubs_path": self._get_api_stubs_path(),
             "write_block_size": self._get_write_block_size(),
+            "write_block_delay": self._get_write_block_delay(),
             "proxy_class": self.__class__.__name__,
         }
         if self._port == "webrepl":
@@ -164,6 +170,9 @@ class BareMetalMicroPythonProxy(MicroPythonProxy):
     def _get_write_block_size(self):
         return get_workbench().get_option(self.backend_name + ".write_block_size")
 
+    def _get_write_block_delay(self):
+        return get_workbench().get_option(self.backend_name + ".write_block_delay")
+
     def interrupt(self):
         # Don't interrupt local process, but direct it to device
         self._send_msg(ImmediateCommand("interrupt"))
@@ -349,6 +358,8 @@ class BareMetalMicroPythonProxy(MicroPythonProxy):
     def open_custom_system_shell(self):
         from thonny import terminal
 
+        get_runner().send_command_and_wait(InlineCommand("prepare_disconnect"), "Disconnecting")
+
         self.disconnect()
 
         terminal.run_in_terminal(
@@ -921,32 +932,32 @@ def add_micropython_backend(
     sort_key=None,
     validate_time=True,
     sync_time=None,
+    utc_clock=False,
     write_block_size=None,
+    write_block_delay=None,
     dtr=None,
     rts=None,
 ):
     if bare_metal:
-        # https://forum.micropython.org/viewtopic.php?f=15&t=3698
-        # https://forum.micropython.org/viewtopic.php?f=15&t=4896&p=28132
-        if write_block_size is None:
-            write_block_size = 255
-
         get_workbench().set_default(name + ".port", "auto")
         get_workbench().set_default(name + ".webrepl_url", DEFAULT_WEBREPL_URL)
         get_workbench().set_default(name + ".webrepl_password", "")
         get_workbench().set_default(name + ".write_block_size", write_block_size)
+        # write_block_delay is used only with "raw" submit_mode
+        get_workbench().set_default(name + ".write_block_delay", write_block_delay)
         get_workbench().set_default(name + ".used_vidpids", set())
         get_workbench().set_default(name + ".dtr", dtr)
         get_workbench().set_default(name + ".rts", rts)
+        get_workbench().set_default(name + ".submit_mode", None)
 
         if sync_time is None:
             sync_time = True
-        get_workbench().set_default(name + ".sync_time", sync_time)
     else:
         if sync_time is None:
             sync_time = False
-        get_workbench().set_default(name + ".sync_time", sync_time)
 
+    get_workbench().set_default(name + ".sync_time", sync_time)
+    get_workbench().set_default(name + ".utc_clock", utc_clock)
     get_workbench().set_default(name + ".validate_time", validate_time)
     get_workbench().add_backend(name, proxy_class, description, config_page, sort_key=sort_key)
 
