@@ -112,73 +112,73 @@ def debug(msg):
 
 class MicroPythonBackend(MainBackend, ABC):
     def __init__(self, clean, args):
-
         self._args = args
         self._prev_time = time.time()
         self._local_cwd = None
         self._cwd = args.get("cwd")
         self._progress_times = {}
-        self._welcome_text_printed = False
         self._welcome_text = None
         self._sys_path = None
-
+        self._epoch_year = None
+        self._builtin_modules = []
         self._api_stubs_path = args.get("api_stubs_path")
+        self._builtins_info = self._fetch_builtins_info()
 
         MainBackend.__init__(self)
         try:
             self._report_time("before prepare")
-            self._prepare(clean)
+            self._process_until_initial_prompt(clean)
+            if self._welcome_text is None:
+                self._welcome_text = self._fetch_welcome_text()
+                self._report_time("got welcome")
+
+            self._prepare_after_soft_reboot(clean)
+
+            if not self._builtin_modules:
+                self._builtin_modules = self._fetch_builtin_modules()
+                logger.debug("Built-in modules: %s", self._builtin_modules)
+
+            self._prepare_rtc()
+            self._send_ready_message()
+            self._report_time("sent ready")
             self.mainloop()
         except ConnectionClosedException as e:
             self._on_connection_closed(e)
         except Exception:
             logger.exception("Crash in backend")
-            traceback.print_exc()
-
-    def _prepare(self, clean):
-        self._process_until_initial_prompt(clean)
-        # Can result either in raw or regular prompt
-
-        if self._welcome_text is None:
-            self._welcome_text = self._fetch_welcome_text()
-            self._report_time("got welcome")
-
-        self._configure_write_blocks()
 
+    def _prepare_after_soft_reboot(self, clean):
         self._report_time("bef preparing helpers")
-        self._prepare_helpers()
+        script = self._get_all_helpers()
+        self._check_perform_just_in_case_gc()
+        self._execute_without_output(script)
         self._report_time("prepared helpers")
 
         self._update_cwd()
         self._report_time("got cwd")
-
-        if self._sys_path is None:
-            self._sys_path = self._fetch_sys_path()
-
-        self._epoch_year = self._fetch_epoch_year()
-        self._check_sync_time()
-        if self._args.get("validate_time"):
-            self._validate_time()
-
-        if self._welcome_text:
-            self._send_ready_message()
-            self._report_time("sent ready")
-
-        self._builtin_modules = self._fetch_builtin_modules()
-        self._builtins_info = self._fetch_builtins_info()
+        self._sys_path = self._fetch_sys_path()
 
         self._report_time("prepared")
         self._check_perform_just_in_case_gc()
         logger.info("Prepared")
 
-    def _configure_write_blocks(self):
-        pass
+    def _prepare_rtc(self):
+        if self._epoch_year is None:
+            self._epoch_year = self._fetch_epoch_year()
+
+        self._check_sync_time()
+        if self._args.get("validate_time"):
+            self._validate_time()
 
     def _check_perform_just_in_case_gc(self):
         if self._connected_to_microbit():
             # May fail to allocate memory without this
             self._perform_gc()
 
+    def _check_sync_time(self):
+        if self._args.get("sync_time"):
+            self._sync_time()
+
     def _perform_gc(self):
         self._execute_without_output(
             dedent(
@@ -190,10 +190,8 @@ class MicroPythonBackend(MainBackend, ABC):
             )
         )
 
-    def _prepare_helpers(self):
-        script = self._get_all_helpers()
-        self._check_perform_just_in_case_gc()
-        self._execute_without_output(script)
+    def _check_prepare(self):
+        pass  # overridden in bare metal
 
     def _get_all_helpers(self):
         # Can't import functions into class context:
@@ -216,7 +214,7 @@ class MicroPythonBackend(MainBackend, ABC):
                     if obj is not None:
                         cls.last_repl_values.append(obj)
                         cls.last_repl_values = cls.last_repl_values[-{num_values_to_keep}:]
-                        print({start_marker!r}, cls.repr(obj), '@', id(obj), {end_marker!r}, sep='')
+                        print({start_marker!r} % id(obj), repr(obj), {end_marker!r}, sep='')
                 
                 @staticmethod
                 def print_mgmt_value(obj):
@@ -226,8 +224,8 @@ class MicroPythonBackend(MainBackend, ABC):
                 def repr(obj):
                     try:
                         s = repr(obj)
-                        if len(s) > 1000:
-                            s = s[:1000] + "..."
+                        if len(s) > 50:
+                            s = s[:50] + "..."
                         return s
                     except Exception as e:
                         return "<could not serialize: " + str(e) + ">"
@@ -258,15 +256,17 @@ class MicroPythonBackend(MainBackend, ABC):
         in internal lists for the purpose of retrieving them by id for Object inspector"""
         return 5
 
-    def _check_sync_time(self):
-        if self._args.get("sync_time"):
-            self._sync_time()
-
     def _sync_time(self):
         raise NotImplementedError()
 
+    def _get_time_for_rtc(self):
+        if self._args["utc_clock"]:
+            return datetime.datetime.now(tz=datetime.timezone.utc).timetuple()
+        else:
+            return datetime.datetime.now().timetuple()
+
     def _validate_time(self):
-        local = datetime.datetime.now(tz=datetime.timezone.utc).timetuple()
+        this_computer = self._get_time_for_rtc()
         remote = self._get_utc_timetuple_from_device()
         if isinstance(remote, tuple):
             # tweak the format if required
@@ -274,7 +274,7 @@ class MicroPythonBackend(MainBackend, ABC):
             while len(remote) < 8:
                 remote += (0,)
             remote += (-1,)  # unknown DST
-            diff = int(time.mktime(local) - time.mktime(remote))
+            diff = int(time.mktime(this_computer) - time.mktime(remote))
             if abs(diff) > 10:
                 print("WARNING: Device's real-time clock seems to be off by %s seconds" % diff)
         else:
@@ -304,10 +304,10 @@ class MicroPythonBackend(MainBackend, ABC):
 
     def _handle_immediate_command(self, cmd: ImmediateCommand) -> None:
         if cmd["name"] == "interrupt":
+            # don't interrupt while command or input is being written
             with self._interrupt_lock:
                 if self._current_command:
                     self._current_command.interrupted = True
-                # don't interrupt while command or input is being written
                 self._write(INTERRUPT_CMD)
                 time.sleep(0.1)
                 self._write(INTERRUPT_CMD)
@@ -315,7 +315,7 @@ class MicroPythonBackend(MainBackend, ABC):
                 self._write(INTERRUPT_CMD)
 
     def _handle_normal_command(self, cmd: CommandToBackend) -> None:
-
+        logger.info("Handling command '%s'", cmd.name)
         self._report_time("before " + cmd.name)
         assert isinstance(cmd, (ToplevelCommand, InlineCommand))
 
@@ -338,10 +338,10 @@ class MicroPythonBackend(MainBackend, ABC):
         else:
             try:
                 response = handler(cmd)
-            except SystemExit:
+            except SystemExit as e:
                 # Must be caused by Thonny or plugins code
                 if isinstance(cmd, ToplevelCommand):
-                    traceback.print_exc()
+                    logger.exception("Unexpected SystemExit", exc_info=e)
                 response = create_error_response(SystemExit=True)
             except UserError as e:
                 sys.stderr.write(str(e) + "\n")
@@ -362,8 +362,8 @@ class MicroPythonBackend(MainBackend, ABC):
                     self._show_error("STDERR:\n" + e.err + "\n")
 
                     response = create_error_response(error="ManagementError")
-            except Exception:
-                _report_internal_error()
+            except Exception as e:
+                _report_internal_error(e)
                 response = create_error_response(context_info="other unhandled exception")
 
         if response is None:
@@ -414,28 +414,17 @@ class MicroPythonBackend(MainBackend, ABC):
         if not self._supports_directories():
             return []
         else:
-            return self._evaluate(
-                dedent(
-                    """
-            import sys as __thonny_sys
-            __thonny_helper.print_mgmt_value(__thonny_sys.path)
-            del __thonny_sys
-            """
-                )
-            )
+            return self._evaluate("__thonny_helper.sys.path")
 
     def _fetch_builtins_info(self):
-        """
-        for p in self._get_api_stubs_path():
-            builtins_file = os.path.join(p, "__builtins__.py")
-            if os.path.exists(builtins_file):
-                return parse_api_information(builtins_file)
-        """
-        path = os.path.join(self._api_stubs_path, "builtins.py")
-        if os.path.exists(path):
-            return parse_api_information(path)
-        else:
-            return {}
+        result = {}
+
+        for name in ["builtins.py", "builtins.pyi"]:
+            path = os.path.join(self._api_stubs_path, name)
+            if os.path.exists(path):
+                result = parse_api_information(path)
+
+        return result
 
     def _fetch_epoch_year(self):
         if self._connected_to_microbit():
@@ -475,14 +464,12 @@ class MicroPythonBackend(MainBackend, ABC):
             return result
 
     def _update_cwd(self):
-        if "micro:bit" not in self._welcome_text.lower():
+        if not self._connected_to_microbit():
             self._cwd = self._evaluate("__thonny_helper.getcwd()")
 
     def _send_ready_message(self):
         args = dict(cwd=self._cwd)
-        # if not clean, then welcome text is already printed as output from the last session
-        if not self._welcome_text_printed:
-            args["welcome_text"] = self._welcome_text
+        args["welcome_text"] = self._welcome_text
 
         self.send_message(ToplevelResponse(**args))
 
@@ -627,15 +614,7 @@ class MicroPythonBackend(MainBackend, ABC):
             raise UserError("%cd takes one parameter")
 
     def _cmd_Run(self, cmd):
-        """Only for %run $EDITOR_CONTENT. start runs will be handled differently."""
-        if cmd.get("source"):
-            self._reset_environment()
-            self._execute(
-                self._avoid_printing_expression_statements(cmd.source), capture_output=False
-            )
-            self._prepare_helpers()
-            self._update_cwd()
-        return {}
+        raise NotImplementedError()
 
     def _cmd_execute_source(self, cmd):
         # TODO: clear last object inspector requests dictionary
@@ -643,9 +622,8 @@ class MicroPythonBackend(MainBackend, ABC):
             source = self._add_expression_statement_handlers(cmd.source)
             self._report_time("befexeccc")
             self._execute(source, capture_output=False)
+            self._check_prepare()
             self._report_time("affexeccc")
-            self._update_cwd()
-            self._report_time("upcd")
         # TODO: assign last value to _
         return {}
 
@@ -655,7 +633,7 @@ class MicroPythonBackend(MainBackend, ABC):
     def _cmd_get_globals(self, cmd):
         if cmd.module_name == "__main__":
             globs = self._evaluate(
-                "{name : __thonny_helper.repr(value) for (name, value) in globals().items() if not name.startswith('__')}"
+                "{name : (__thonny_helper.repr(value), id(value)) for (name, value) in globals().items() if not name.startswith('__')}"
             )
         else:
             globs = self._evaluate(
@@ -663,7 +641,8 @@ class MicroPythonBackend(MainBackend, ABC):
                     """
                 import %s as __mod_for_globs
                 __thonny_helper.print_mgmt_value(
-                    {name : __thonny_helper.repr(getattr(__mod_for_globs, name)) 
+                    {name : (__thonny_helper.repr(getattr(__mod_for_globs, name)), 
+                             id(getattr(__mod_for_globs, name)))
                         in dir(__mod_for_globs) 
                         if not name.startswith('__')}
                 )
@@ -671,7 +650,12 @@ class MicroPythonBackend(MainBackend, ABC):
             """
                 )
             )
-        return {"module_name": cmd.module_name, "globals": globs}
+
+        value_infos = {}
+        for name, pair in globs.items():
+            value_infos[name] = ValueInfo(pair[1], pair[0])
+
+        return {"module_name": cmd.module_name, "globals": value_infos}
 
     def _cmd_get_fs_info(self, cmd):
         raise NotImplementedError()
@@ -697,51 +681,41 @@ class MicroPythonBackend(MainBackend, ABC):
 
         # need to keep the reference corresponding to object_id so that it can be later found as next context object
         # remove non-relevant items
-        relevant = set([cmd.object_id] + cmd.back_links + cmd.forward_links)
+        # TODO: add back links
+        # relevant = set([cmd.object_id] + cmd.back_links + cmd.forward_links)
         self._execute(
             dedent(
                 """
-                if id(__thonny_value) not in __thonny_helper.inspector_values:
-                    __thonny_helper.inspector_values[id(__thonny_value)] = __thonny_value
-                __thonny_relevant = %r
-                #__thonny_helper.inspector_values = {
-                #    obj_id : __thonny_helper.inspector_values[obj_id] for obj_id in __thonny_helper.inspector_values
-                #    if obj_id in __thonny_relevant
-                #}
+                if id(__thonny_helper.object_info) not in __thonny_helper.inspector_values:
+                    __thonny_helper.inspector_values[id(__thonny_helper.object_info)] = __thonny_helper.object_info
             """
-                % relevant
             )
         )
 
-        self._execute("del __thonny_value")
-
         return {"id": cmd.object_id, "info": info}
 
-    def _reset_environment(self):
-        """For preparing a Run"""
-        # TODO: clear last object inspector requests dictionary
-        pass
-
     def _find_basic_object_info(self, object_id, context_id):
-        """If object is found then returns basic info and leaves object reference to __thonny_value"""
+        """If object is found then returns basic info and leaves object reference
+        to __thonny_helper.object_info.
+
+        Can't leave it in a global object, because when querying globals(),
+        repr(globals()) would cause inifite recursion."""
 
         result = self._evaluate(
             dedent(
                 """
-                for __thonny_value in (
+                for __thonny_helper.object_info in (
                         list(globals().values()) 
                         + __thonny_helper.last_repl_values
                         + list(__thonny_helper.inspector_values.values())):
-                    #print("testing", id(__thonny_value), __thonny_value)
-                    if id(__thonny_value) == %d:
+                    if id(__thonny_helper.object_info) == %d:
                         __thonny_helper.print_mgmt_value({
-                            "repr" : __thonny_helper.repr(__thonny_value),
-                            "type": str(type(__thonny_value))
+                            "repr" : repr(__thonny_helper.object_info),
+                            "type": str(type(__thonny_helper.object_info))
                         })
                         break
                 else:
-                    __thonny_value = None
-                    del __thonny_value
+                    __thonny_helper.object_info = None
                     __thonny_helper.print_mgmt_value(None)
             """
                 % object_id
@@ -754,34 +728,33 @@ class MicroPythonBackend(MainBackend, ABC):
             return self._evaluate(
                 dedent(
                     """
+                __thonny_helper.context_value = __thonny_helper.inspector_values.get(%d, None)
                 
-                __thonny_context_value = __thonny_helper.inspector_values.get(%d, None)
-                
-                if __thonny_context_value is None:
-                    __thonny_value = None
+                if __thonny_helper.context_value is None:
+                    __thonny_helper.object_info = None
                     __thonny_helper.print_mgmt_value(None)
                 else:
-                    __thonny_children = [getattr(__thonny_context_value, name) for name in dir(__thonny_context_value)]
-                    if isinstance(__thonny_context_value, (set, tuple, list)):
-                        __thonny_children += list(__thonny_context_value)
-                    elif isinstance(__thonny_context_value, dict):
-                        __thonny_children += list(__thonny_context_value.values())
+                    __thonny_helper.context_children = [
+                         getattr(__thonny_helper.context_value, name)
+                         for name in dir(__thonny_helper.context_value)
+                    ]
+                    if isinstance(__thonny_helper.context_value, (set, tuple, list)):
+                        __thonny_helper.context_children += list(__thonny_helper.context_value)
+                    elif isinstance(__thonny_helper.context_value, dict):
+                        __thonny_helper.context_children += list(__thonny_helper.context_value.values())
                     
-                    for __thonny_value in __thonny_children:
-                        if id(__thonny_value) == %d:
+                    for __thonny_helper.object_info in __thonny_helper.context_children:
+                        if id(__thonny_helper.object_info) == %d:
                             __thonny_helper.print_mgmt_value({
-                                "repr" : __thonny_helper.repr(__thonny_value),
-                                "type": str(type(__thonny_value))
+                                "repr" : __thonny_helper.repr(__thonny_helper.object_info),
+                                "type": str(type(__thonny_helper.object_info))
                             })
                             break
                     else:
-                        __thonny_value = None
-                        del __thonny_value
-                        __thonny_helper.print_mgmt_value(None)
+                        __thonny_helper.object_info = None
                         
-                    del __thonny_children
-                
-                del __thonny_context_value
+                __thonny_helper.context_value = None
+                __thonny_helper.context_children = None
             """
                     % (context_id, object_id)
                 )
@@ -790,9 +763,12 @@ class MicroPythonBackend(MainBackend, ABC):
             return None
 
     def _get_object_attributes(self, all_attributes):
-        """object is given in __thonny_value """
+        """object is given in __thonny_helper.object_info """
         atts = self._evaluate(
-            "{name : (id(getattr(__thonny_value, name)), __thonny_helper.repr(getattr(__thonny_value, name))) for name in dir(__thonny_value)}"
+            "{name : ("
+            "   id(getattr(__thonny_helper.object_info, name)),"
+            "    __thonny_helper.repr(getattr(__thonny_helper.object_info, name))"
+            ") for name in dir(__thonny_helper.object_info)}"
         )
         return {
             name: ValueInfo(atts[name][0], atts[name][1])
@@ -801,13 +777,16 @@ class MicroPythonBackend(MainBackend, ABC):
         }
 
     def _get_object_info_extras(self, type_name):
-        """object is given in __thonny_value """
+        """object is given in __thonny_helper.object_info """
         if type_name in ("list", "tuple", "set"):
-            items = self._evaluate("[(id(x), __thonny_helper.repr(x)) for x in __thonny_value]")
+            items = self._evaluate(
+                "[(id(x), __thonny_helper.repr(x)) for x in __thonny_helper.object_info]"
+            )
             return {"elements": [ValueInfo(x[0], x[1]) for x in items]}
         elif type_name == "dict":
             items = self._evaluate(
-                "[((id(key), __thonny_helper.repr(key)), (id(__thonny_value[key]), __thonny_helper.repr(__thonny_value[key]))) for key in __thonny_value]"
+                "[((id(key), __thonny_helper.repr(key)), (id(__thonny_helper.object_info[key]), "
+                "__thonny_helper.repr(__thonny_helper.object_info[key]))) for key in __thonny_helper.object_info]"
             )
             return {
                 "entries": [
@@ -973,13 +952,18 @@ class MicroPythonBackend(MainBackend, ABC):
         result = dict(source=cmd.source, row=cmd.row, column=cmd.column)
 
         try:
-            import jedi
+            from thonny import jedi_utils
 
-            script = jedi.Script(cmd.source, cmd.row, cmd.column, sys_path=[self._api_stubs_path])
-            completions = script.completions()
+            completions = jedi_utils.get_script_completions(
+                cmd.source,
+                cmd.row,
+                cmd.column,
+                filename=cmd.filename,
+                sys_path=[self._api_stubs_path],
+            )
             result["completions"] = self._filter_completions(completions)
-        except Exception:
-            traceback.print_exc()
+        except Exception as e:
+            logger.exception("Problem with editor autocomplete", exc_info=e)
             result["error"] = "Autocomplete error"
 
         return result
@@ -1016,17 +1000,16 @@ class MicroPythonBackend(MainBackend, ABC):
             response = {"source": cmd.source}
 
             try:
-                import jedi
+                from thonny import jedi_utils
 
                 # at the moment I'm assuming source is the code before cursor, not whole input
                 lines = source.split("\n")
-                script = jedi.Script(
-                    source, len(lines), len(lines[-1]), sys_path=[self._api_stubs_path]
+                completions = jedi_utils.get_script_completions(
+                    source, len(lines), len(lines[-1]), "<shell>", sys_path=[self._api_stubs_path]
                 )
-                completions = script.completions()
                 response["completions"] = self._filter_completions(completions)
-            except Exception:
-                traceback.print_exc()
+            except Exception as e:
+                logger.exception("Problem with shell autocomplete", exc_info=e)
                 response["error"] = "Autocomplete error"
 
             return response
@@ -1264,8 +1247,10 @@ class MicroPythonBackend(MainBackend, ABC):
             # make sure it parses
             ast.parse(new_source)
             return new_source
-        except Exception:
-            logger.exception("Problem adding Expr handlers")
+        except SyntaxError:
+            return source
+        except Exception as e:
+            logger.warning("Problem adding Expr handlers", exc_info=e)
             return source
 
     def _avoid_printing_expression_statements(self, source):
@@ -1303,8 +1288,10 @@ class MicroPythonBackend(MainBackend, ABC):
             # make sure it parses
             ast.parse(new_source)
             return new_source
-        except Exception:
-            logger.exception("Problem adding Expr handlers")
+        except SyntaxError:
+            return source
+        except Exception as e:
+            logger.warning("Problem adding Expr handlers", exc_info=e)
             return source
 
     def _report_time(self, caption):
@@ -1313,7 +1300,12 @@ class MicroPythonBackend(MainBackend, ABC):
         self._prev_time = new_time
 
     def _system_time_to_posix_time(self, value: float) -> float:
-        return value + self._get_epoch_offset()
+        result = value + self._get_epoch_offset()
+        if not self._args["utc_clock"]:
+            # convert to UTC
+            result += time.timezone
+
+        return result
 
     def _get_epoch_offset(self) -> int:
         if self._epoch_year == 1970:
@@ -1381,9 +1373,8 @@ class ManagementError(RuntimeError):
         self.err = err
 
 
-def _report_internal_error():
-    print("PROBLEM WITH THONNY'S BACK-END:\n", file=sys.stderr)
-    traceback.print_exc()
+def _report_internal_error(exception=None):
+    logger.exception("PROBLEM WITH THONNY'S BACK-END:", exc_info=exception)
 
 
 def parse_api_information(file_path):
