@@ -38,6 +38,14 @@ from thonny.plugins.micropython.webrepl_connection import (
     WebreplBinaryMsg,
 )
 
+PASTE_SUBMIT_MODE = "paste"
+RAW_PASTE_SUBMIT_MODE = "raw_paste"
+RAW_SUBMIT_MODE = "raw"
+
+RAW_PASTE_COMMAND = b"\x05A\x01"
+RAW_PASTE_CONFIRMATION = b"R\x01"
+RAW_PASTE_CONTINUE = b"\x01"
+
 BAUDRATE = 115200
 ENCODING = "utf-8"
 
@@ -119,10 +127,40 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
         self._connection = connection
         self._startup_time = time.time()
         self._interrupt_suggestion_given = False
-        self._write_block_size = args.get("write_block_size", 255)
+
+        # https://forum.micropython.org/viewtopic.php?f=15&t=3698
+        # https://forum.micropython.org/viewtopic.php?f=15&t=4896&p=28132
+        self._write_block_size = args.get("write_block_size", None)
+        self._write_block_delay = args.get("write_block_delay", None)
+        if self._write_block_size is None:
+            self._write_block_size = 255
+        if self._write_block_delay is None:
+            if self._connected_over_webrepl():
+                # ESP-32 needs long delay to work reliably over raw mode WebREPL
+                # TODO: consider removing when this gets fixed
+                self._write_block_delay = 0.5
+            else:
+                self._write_block_delay = 0.01
+
+        self._submit_mode = args.get("submit_mode", None)
+        logger.debug(
+            "Initial submit_mode: %s, block_size: %s, block_delay: %s",
+            self._submit_mode,
+            self._write_block_size,
+            self._write_block_delay,
+        )
+
+        self._last_prompt = None
 
         MicroPythonBackend.__init__(self, clean, args)
 
+    def _check_prepare(self):
+        out, err = self._execute("print('__thonny_helper' in dir())", capture_output=True)
+        if out.strip() == "True":
+            return
+
+        self._prepare_after_soft_reboot(False)
+
     def _get_custom_helpers(self):
         if self._connected_to_microbit():
             return ""
@@ -153,21 +191,46 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
             self._interrupt_to_raw_prompt()
             self._soft_reboot_in_raw_prompt_without_running_main()
         else:
-            # Order first normal prompt to be output when the code is done.
-            # If the device is already at normal prompt then it gets repeated (together with the banner)
-            # If it is at raw prompt then outputs full normal prompt.
-            # If it is in the middle of input, then it is ignored.
-            self._write(NORMAL_MODE_CMD)
+            # Discard what's printed by now and order a prompt, so that we get to know
+            # if the REPL is already idle
+            discarded = self._connection.read_all()
+            self._write(RAW_MODE_CMD)
             self._forward_output_until_active_prompt(self._send_output)
-            self._welcome_text_printed = True
+
+        if self._submit_mode is None:
+            self._choose_submit_mode()
+
+    def _choose_submit_mode(self):
+        if self._connected_over_webrepl():
+            logger.info("Choosing paste submit mode because of WebREPL")
+            self._submit_mode = PASTE_SUBMIT_MODE
+            return
+
+        # at least sometimes, we end up at normal prompt, although we asked for raw prompt
+        self._ensure_raw_mode()
+        self._write(RAW_PASTE_COMMAND)
+        response = self._connection.soft_read(2)
+        assert len(response) == 2, "Could not read response for raw paste command: " + response
+        if response == RAW_PASTE_CONFIRMATION:
+            logger.info("Choosing raw paste submit mode")
+            self._submit_mode = RAW_PASTE_SUBMIT_MODE
+            self._write(EOT)
+            discarding = self._connection.read_until(RAW_PROMPT)
+        else:
+            discarding = self._connection.read_until(RAW_PROMPT)
+            logger.info("Choosing raw submit mode (%r)", response + discarding)
+            self._submit_mode = RAW_SUBMIT_MODE
+
+        discarding += self._connection.read_all()
 
     def _fetch_welcome_text(self) -> str:
         self._write(NORMAL_MODE_CMD)
-        welcome_text = self._connection.read_until(NORMAL_PROMPT).strip(b"\r\n >")
+        out, err = self._capture_output_until_active_prompt()
+        welcome_text = out.strip("\r\n >")
         if os.name != "nt":
-            welcome_text = welcome_text.replace(b"\r\n", b"\n")
+            welcome_text = welcome_text.replace("\r\n", "\n")
 
-        return self._decode(welcome_text)
+        return welcome_text
 
     def _fetch_builtin_modules(self):
         script = "help('modules')"
@@ -202,10 +265,9 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
             return 2000
 
     def _sync_time(self):
-        """Sets the time on the pyboard to match the time on the host."""
+        """Sets the time to match the time on the host."""
 
-        # RTC works on UTC
-        now = datetime.datetime.now(tz=datetime.timezone.utc).timetuple()
+        now = self._get_time_for_rtc()
 
         if self._connected_to_microbit():
             return
@@ -346,10 +408,8 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
             self._write(INTERRUPT_CMD)
             self._write(RAW_MODE_CMD)
             time.sleep(delay)
-            discarded_bytes += self._connection.read_all()
-            if discarded_bytes.endswith(FIRST_RAW_PROMPT) or discarded_bytes.endswith(
-                W600_FIRST_RAW_PROMPT
-            ):
+            self._capture_output_until_active_prompt()
+            if self._last_prompt in [FIRST_RAW_PROMPT, W600_FIRST_RAW_PROMPT]:
                 break
         else:
             max_tail_length = 500
@@ -378,68 +438,74 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
         logger.debug("_soft_reboot_in_raw_prompt_without_running_main")
         self._write(SOFT_REBOOT_CMD + INTERRUPT_CMD)
         self._check_reconnect()
-        self._extra_interrupts_after_soft_reboot()
-
-        discarded_bytes = b""
-        while not discarded_bytes.endswith(FIRST_RAW_PROMPT) or discarded_bytes.endswith(
-            W600_FIRST_RAW_PROMPT
-        ):
-            discarded_bytes += self._connection.soft_read(1, timeout=1)
-            discarded_bytes += self._connection.read_all()
-            logger.debug("Discarded bytes when waiting for reboot: %r", discarded_bytes)
+        self._capture_output_until_active_prompt()
 
         logger.debug("Done soft reboot in raw prompt")
 
-    def _enter_raw_mode(self):
-        logger.debug("_enter_raw_mode")
+    def _ensure_raw_mode(self):
+        if self._last_prompt in [
+            RAW_PROMPT,
+            EOT + RAW_PROMPT,
+            FIRST_RAW_PROMPT,
+            W600_FIRST_RAW_PROMPT,
+        ]:
+            return
+        logger.debug("requesting raw mode at %r", self._last_prompt)
+
         # assuming we are currently on a normal prompt
         self._write(RAW_MODE_CMD)
-        discarded = self._connection.read_until(b"\r\n>", timeout=2)
-        discarded += self._connection.read_all()
-        if not discarded.endswith(FIRST_RAW_PROMPT) and not discarded.endswith(
-            W600_FIRST_RAW_PROMPT
-        ):
-            raise AssertionError("Could not enter raw prompt, got " + repr(discarded))
+        out, err = self._capture_output_until_active_prompt()
+        if self._last_prompt == NORMAL_PROMPT:
+            # Don't know why this happens sometimes (eg. when interrupting a Ctrl+D or restarted
+            # program, which is outputting text on ESP32)
+            logger.info("Found normal prompt instead of expected raw prompt. Trying again.")
+            self._write(RAW_MODE_CMD)
+            time.sleep(0.5)
+            out, err = self._capture_output_until_active_prompt()
 
-    def _enter_normal_mode(self):
-        logger.debug("_enter_normal_mode")
+        if self._last_prompt not in [FIRST_RAW_PROMPT, W600_FIRST_RAW_PROMPT]:
+            raise AssertionError(
+                "Could not enter raw prompt, got %r"
+                % ((out + err).encode(ENCODING) + self._last_prompt)
+            )
+
+    def _ensure_normal_mode(self):
+        if self._last_prompt == NORMAL_PROMPT:
+            return
+
+        logger.debug("requesting normal mode at %r", self._last_prompt)
         self._write(NORMAL_MODE_CMD)
+        self._capture_output_until_active_prompt()
+        assert self._last_prompt == NORMAL_PROMPT, (
+            "Could not get normal prompt, got %s" % self._last_prompt
+        )
 
-        def ignore_output(data, stream):
-            pass
-
-        self._forward_output_until_active_prompt(ignore_output)
-
-    def _soft_reboot_in_normal_mode_without_running_main(self):
-        logger.debug("_soft_reboot_in_normal_mode_without_running_main")
-        self._enter_raw_mode()
+    def _soft_reboot_without_running_main(self):
+        logger.debug("_soft_reboot_without_running_main")
+        self._interrupt_to_raw_prompt()
         self._soft_reboot_in_raw_prompt_without_running_main()
-        self._enter_normal_mode()
-
-    def _extra_interrupts_after_soft_reboot(self):
-        # To be overridden for CP
-        pass
 
     def _soft_reboot_for_restarting_user_program(self):
         # Need to go to normal mode. MP doesn't run user code in raw mode
         # (CP does, but it doesn't hurt to do it there as well)
-        self._write(NORMAL_MODE_CMD)
-        self._connection.read_until(NORMAL_PROMPT)
+        self._ensure_normal_mode()
         self._write(SOFT_REBOOT_CMD)
         self._check_reconnect()
         self._forward_output_until_active_prompt(self._send_output)
         logger.debug("Restoring helpers")
-        self._prepare_helpers()
-        self._update_cwd()
+        self._prepare_after_soft_reboot(False)
         self.send_message(ToplevelResponse(cwd=self._cwd))
 
     def _check_reconnect(self):
-        from thonny.plugins.micropython.webrepl_connection import WebReplConnection
-
-        if isinstance(self._connection, WebReplConnection):
+        if self._connected_over_webrepl():
             time.sleep(1)
             self._connection = self._connection.close_and_return_new_connection()
 
+    def _connected_over_webrepl(self):
+        from thonny.plugins.micropython.webrepl_connection import WebReplConnection
+
+        return isinstance(self._connection, WebReplConnection)
+
     def _transform_output(self, data, stream_name):
         # Any keypress wouldn't work
         return data.replace(
@@ -460,17 +526,15 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
             cdata = cdata[:-1] + "\r\n"
 
         bdata = cdata.encode(ENCODING)
-
+        to_be_written = bdata
+        echo = b""
         with self._interrupt_lock:
-            self._write(bdata)
-            # Try to consume the echo
-
-            try:
-                echo = self._connection.read(len(bdata))
-            except queue.Empty:
-                # leave it.
-                logging.warning("Timeout when reading input echo")
-                return
+            while to_be_written:
+                block = self._extract_block_without_splitting_chars(to_be_written)
+                self._write(to_be_written)
+                # Try to consume the echo
+                echo += self._connection.soft_read(len(block), timeout=1)
+                to_be_written = to_be_written[len(block) :]
 
         if echo != bdata:
             # because of autoreload? timing problems? interruption?
@@ -490,64 +554,160 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
         """
         assert script
 
-        # assuming we are already in a prompt, but threads may have produced something
+        # assuming we are already at a prompt, but threads may have produced something
         self._forward_unexpected_output()
 
         to_be_sent = script.encode("UTF-8")
+        logger.debug("Submitting via %s: %r", self._submit_mode, to_be_sent[:50])
         with self._interrupt_lock:
-            # Go to paste mode
-            self._connection.write(PASTE_MODE_CMD)
-            self._connection.read_until(PASTE_MODE_LINE_PREFIX)
+            if self._submit_mode == PASTE_SUBMIT_MODE:
+                self._submit_code_via_paste_mode(to_be_sent)
+            elif self._submit_mode == RAW_PASTE_SUBMIT_MODE:
+                try:
+                    self._submit_code_via_raw_paste_mode(to_be_sent)
+                except RawPasteNotSupportedError:
+                    logger.info("WARNING: Could not use expected raw paste, falling back to raw")
+                    self._submit_code_via_raw_mode(to_be_sent)
+            else:
+                self._submit_code_via_raw_mode(to_be_sent)
 
-            # Send script
-            while to_be_sent:
-                block = to_be_sent[: self._write_block_size]
-                to_be_sent = to_be_sent[self._write_block_size :]
+    def _submit_code_via_paste_mode(self, script_bytes: bytes) -> None:
+        # Go to paste mode
+        self._ensure_normal_mode()
+        self._connection.write(PASTE_MODE_CMD)
+        self._connection.read_until(PASTE_MODE_LINE_PREFIX)
 
-                # find proper block boundary
-                while True:
-                    expected_echo = block.replace(b"\r\n", b"\r\n" + PASTE_MODE_LINE_PREFIX)
-                    if (
-                        len(expected_echo) > self._write_block_size
-                        or block.endswith(b"\r")
-                        or len(block) > 2
-                        and starts_with_continuation_byte(to_be_sent)
-                    ):
-                        # move last byte to the next block
-                        to_be_sent = block[-1:] + to_be_sent
-                        block = block[:-1]
-                        continue
-                    else:
-                        break
+        # Send script
+        while script_bytes:
+            block = script_bytes[: self._write_block_size]
+            script_bytes = script_bytes[self._write_block_size :]
 
-                self._write(block)
-                self._connection.read_all_expected(expected_echo, timeout=WAIT_OR_CRASH_TIMEOUT)
+            # find proper block boundary
+            while True:
+                expected_echo = block.replace(b"\r\n", b"\r\n" + PASTE_MODE_LINE_PREFIX)
+                if (
+                    len(expected_echo) > self._write_block_size
+                    or block.endswith(b"\r")
+                    or len(block) > 2
+                    and starts_with_continuation_byte(script_bytes)
+                ):
+                    # move last byte to the next block
+                    script_bytes = block[-1:] + script_bytes
+                    block = block[:-1]
+                    continue
+                else:
+                    break
 
-            # push and read comfirmation
-            self._connection.write(EOT)
-            expected_confirmation = b"\r\n"
-            actual_confirmation = self._connection.read(
-                len(expected_confirmation), timeout=WAIT_OR_CRASH_TIMEOUT
-            )
-            assert actual_confirmation == expected_confirmation, "Expected %r, got %r" % (
-                expected_confirmation,
-                actual_confirmation,
+            self._write(block)
+            self._connection.read_all_expected(expected_echo, timeout=WAIT_OR_CRASH_TIMEOUT)
+
+        # push and read comfirmation
+        self._connection.write(EOT)
+        expected_confirmation = b"\r\n"
+        actual_confirmation = self._connection.read(
+            len(expected_confirmation), timeout=WAIT_OR_CRASH_TIMEOUT
+        )
+        assert actual_confirmation == expected_confirmation, "Expected %r, got %r" % (
+            expected_confirmation,
+            actual_confirmation,
+        )
+
+    def _submit_code_via_raw_mode(self, script_bytes: bytes) -> None:
+        self._ensure_raw_mode()
+
+        to_be_written = script_bytes + EOT
+
+        while to_be_written:
+            block = self._extract_block_without_splitting_chars(to_be_written)
+            self._write(block)
+            to_be_written = to_be_written[len(block) :]
+            if to_be_written:
+                time.sleep(self._write_block_delay)
+
+        # fetch command confirmation
+        confirmation = self._connection.soft_read(2, timeout=WAIT_OR_CRASH_TIMEOUT)
+
+        if confirmation != OK:
+            data = confirmation + self._connection.read_all()
+            data += self._connection.read(1, timeout=1, timeout_is_soft=True)
+            data += self._connection.read_all()
+            raise AssertionError(
+                "Could not read command confirmation. Got "
+                + repr(data)
+                + "\n\nSCRIPT:\n"
+                + repr(script_bytes)
             )
 
+    def _submit_code_via_raw_paste_mode(self, script_bytes: bytes) -> None:
+        self._ensure_raw_mode()
+        self._connection.set_unicode_guard(False)
+        self._write(RAW_PASTE_COMMAND)
+        response = self._connection.soft_read(2, timeout=WAIT_OR_CRASH_TIMEOUT)
+        if response != RAW_PASTE_CONFIRMATION:
+            # Occasionally, the device initially supports raw paste but later doesn't allow it
+            # https://github.com/thonny/thonny/issues/1545
+            time.sleep(0.01)
+            response += self._connection.read_all()
+            if response == FIRST_RAW_PROMPT:
+                self._last_prompt = FIRST_RAW_PROMPT
+                raise RawPasteNotSupportedError()
+            else:
+                raise AssertionError("Got %r instead of raw-paste confirmation" % response)
+
+        self._raw_paste_write(script_bytes)
+        self._connection.set_unicode_guard(True)
+
+    def _raw_paste_write(self, command_bytes):
+        # Adapted from https://github.com/micropython/micropython/commit/a59282b9bfb6928cd68b696258c0dd2280244eb3#diff-cf10d3c1fe676599a983c0ec85b78c56c9a6f21b2d896c69b3e13f34d454153e
+
+        # Read initial header, with window size.
+        data = self._connection.soft_read(2, timeout=2)
+        assert len(data) == 2, "Could not read initial header, got %r" % (
+            data + self._connection.read_all()
+        )
+        window_size = data[0] | data[1] << 8
+        window_remain = window_size
+
+        # Write out the command_bytes data.
+        i = 0
+        while i < len(command_bytes):
+            while window_remain == 0 or not self._connection.incoming_is_empty():
+                data = self._connection.soft_read(1, timeout=WAIT_OR_CRASH_TIMEOUT)
+                if data == b"\x01":
+                    # Device indicated that a new window of data can be sent.
+                    window_remain += window_size
+                elif data == b"\x04":
+                    # Device indicated abrupt end, most likely a syntax error.
+                    # Acknowledge it and finish.
+                    self._connection.write(b"\x04")
+                    logger.debug(
+                        "Abrupt end of raw paste submit after submitting %s bytes out of %s",
+                        i,
+                        len(command_bytes),
+                    )
+                    return
+                else:
+                    # Unexpected data from device.
+                    raise AssertionError("Unexpected read during raw paste: {}".format(data))
+            # Send out as much data as possible that fits within the allowed window.
+            b = command_bytes[i : min(i + window_remain, len(command_bytes))]
+            self._write(b)
+            window_remain -= len(b)
+            i += len(b)
+
+        # Indicate end of data.
+        self._write(b"\x04")
+
+        # Wait for device to acknowledge end of data.
+        data = self._connection.soft_read_until(b"\x04", timeout=WAIT_OR_CRASH_TIMEOUT)
+        if not data.endswith(b"\x04"):
+            raise AssertionError("could not complete raw paste: {}".format(data))
+
     def _execute_with_consumer(self, script, output_consumer: Callable[[str, str], None]):
-        """Expected output after submitting the command and reading the confirmation is following:
-
-        stdout
-        EOT
-        stderr
-        EOT
-        RAW_PROMPT
-        """
-
         self._report_time("befsubcode")
         self._submit_code(script)
         self._report_time("affsubcode")
-        self._forward_output_until_active_prompt(output_consumer, "stdout")
+        self._forward_output_until_active_prompt(output_consumer)
         self._report_time("affforw")
 
     def _forward_output_until_active_prompt(
@@ -602,7 +762,7 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
 
         Because ot the special role of EOT and NORMAL_PROMT, we assume user code
         will not output these. If it does, processing may break.
-        It may succceed if the propmt is followed by something (quickly enough)
+        It may succeed if the prompt is followed by something (quickly enough)
         -- that's why we look for *active* prompt, ie. prompt without following text.
         TODO: Experiment with this!
 
@@ -611,10 +771,20 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
         is from another thread or the main thread is still running.
         For now I'm ignoring these problems and assume all output comes from the main thread.
         """
+
+        # Don't want to block on lone EOT (the first EOT), because finding the second EOT
+        # together with raw prompt marker is the most important.
         INCREMENTAL_OUTPUT_BLOCK_CLOSERS = re.compile(
-            b"|".join(map(re.escape, [NORMAL_PROMPT, LF]))
+            b"|".join(
+                map(
+                    re.escape,
+                    [NORMAL_PROMPT, LF, EOT + RAW_PROMPT, FIRST_RAW_PROMPT, W600_FIRST_RAW_PROMPT],
+                )
+            )
         )
 
+        prompts = [EOT + RAW_PROMPT, NORMAL_PROMPT, FIRST_RAW_PROMPT, W600_FIRST_RAW_PROMPT]
+
         pending = b""
         while True:
             # There may be an input submission waiting
@@ -625,7 +795,23 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
             new_data = self._connection.soft_read_until(
                 INCREMENTAL_OUTPUT_BLOCK_CLOSERS, timeout=0.05
             )
-            if TRACEBACK_MARKER in new_data:
+
+            # Try to separate stderr from stdout in raw mode
+            eot_pos = new_data.find(EOT)
+            if (
+                eot_pos >= 0
+                and new_data[eot_pos : eot_pos + 2] != EOT + RAW_PROMPT
+                and stream_name == "stdout"
+            ):
+                # start of stderr in raw mode
+                out, err = new_data.split(EOT, maxsplit=1)
+                pending += out
+                output_consumer(self._decode(pending), stream_name)
+                pending = b""
+                new_data = err
+                stream_name = "stderr"
+            elif self._submit_mode == PASTE_SUBMIT_MODE and TRACEBACK_MARKER in new_data:
+                # start of stderr in paste mode
                 stream_name = "stderr"
 
             if not new_data:
@@ -651,11 +837,13 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
 
             pending += new_data
 
-            if pending.endswith(LF):
-                output_consumer(self._decode(pending), stream_name)
-                pending = b""
+            for current_prompt in prompts:
+                if pending.endswith(current_prompt):
+                    break
+            else:
+                current_prompt = None
 
-            elif pending.endswith(NORMAL_PROMPT):
+            if current_prompt:
                 # This looks like prompt.
                 # Make sure it is not followed by anything.
                 follow_up = self._connection.soft_read(1, timeout=0.01)
@@ -666,70 +854,113 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
                     # Don't output yet, because the follow up may turn into another prompt
                     # and they can be captured all together.
                     self._connection.unread(follow_up)
-                    # read propmt must remain in pending
+                    # read prompt must remain in pending
+                    continue
                 else:
                     # let's hope it is an active prompt
-                    terminator = NORMAL_PROMPT
-
                     # Strip all trailing prompts
-                    out = pending
                     while True:
-                        if out.endswith(NORMAL_PROMPT):
-                            out = out[: -len(NORMAL_PROMPT)]
+                        for potential_prompt in prompts:
+                            if pending.endswith(potential_prompt):
+                                pending = pending[: -len(potential_prompt)]
                         else:
                             break
-                    output_consumer(self._decode(out), stream_name)
-
-                    return terminator
-
-            elif ends_overlap(pending, NORMAL_PROMPT):
-                # Maybe we have a prefix of the prompt and the rest is still coming?
-                # (it's OK to wait a bit, as the user output usually ends with a newline, ie not
-                # with a prompt prefix)
-                follow_up = self._connection.soft_read(1, timeout=0.3)
-                if not follow_up:
-                    # most likely not a Python prompt, let's forget about it
                     output_consumer(self._decode(pending), stream_name)
+                    self._last_prompt = current_prompt
+                    # logger.debug("Found prompt %r", current_prompt)
+                    return current_prompt
+
+            if pending.endswith(LF):
+                # Maybe it's a penultimate char in a first raw repl?
+                if pending.endswith(FIRST_RAW_PROMPT[:-1]) or pending.endswith(
+                    W600_FIRST_RAW_PROMPT[:-1]
+                ):
+                    pending += self._connection.soft_read(1)
+                    self._connection.unread(pending)
                     pending = b""
                 else:
-                    # Let's try the possible prefix again in the next iteration
-                    # (I'm unreading otherwise the read_until won't see the whole prompt
-                    # and needs to wait for the timeout)
-                    n = ends_overlap(pending, NORMAL_PROMPT)
+                    output_consumer(self._decode(pending), stream_name)
+                    pending = b""
+                continue
 
-                    try_again = pending[-n:]
-                    pending = pending[:-n]
-                    self._connection.unread(try_again + follow_up)
+            for potential_prompt in prompts:
+                if ends_overlap(pending, potential_prompt):
+                    # Maybe we have a prefix of the prompt and the rest is still coming?
+                    # (it's OK to wait a bit, as the user output usually ends with a newline, ie not
+                    # with a prompt prefix)
+                    follow_up = self._connection.soft_read(1, timeout=0.3)
+                    if not follow_up:
+                        # most likely not a Python prompt, let's forget about it
+                        output_consumer(self._decode(pending), stream_name)
+                        pending = b""
+                        continue
+                    else:
+                        # Let's try the possible prefix again in the next iteration
+                        # (I'm unreading otherwise the read_until won't see the whole prompt
+                        # and needs to wait for the timeout)
+                        n = ends_overlap(pending, potential_prompt)
+
+                        try_again = pending[-n:]
+                        pending = pending[:-n]
+                        self._connection.unread(try_again + follow_up)
+                        continue
 
             else:
-                # No EOT or prompt in sight.
+                # No prompt in sight.
                 # Output and keep working.
                 output_consumer(self._decode(pending), stream_name)
                 pending = b""
+                continue
+
+    def _capture_output_until_active_prompt(self):
+        output = {"stdout": "", "stderr": ""}
+
+        def collect_output(data, stream):
+            output[stream] += data
+
+        self._forward_output_until_active_prompt(collect_output)
+
+        return output["stdout"], output["stderr"]
 
     def _forward_unexpected_output(self, stream_name="stdout"):
         "Invoked between commands"
+        # TODO: This should be as careful as _forward_output_until_active_prompt
         data = self._connection.read_all(check_error=False)
         if data:
             met_prompt = False
             while data.endswith(NORMAL_PROMPT) or data.endswith(FIRST_RAW_PROMPT):
                 # looks like the device was resetted
-                met_prompt = True
-
                 if data.endswith(NORMAL_PROMPT):
-                    terminator = NORMAL_PROMPT
+                    prompt = NORMAL_PROMPT
                 else:
-                    terminator = FIRST_RAW_PROMPT
+                    prompt = FIRST_RAW_PROMPT
+
+                if not met_prompt:
+                    met_prompt = True
+                    self._last_prompt = prompt
 
                 # hide the prompt from the output ...
-                data = data[: -len(terminator)]
+                data = data[: -len(prompt)]
 
             self._send_output(data.decode(ENCODING, "replace"), stream_name)
             if met_prompt:
-                self._prepare_helpers()
+                self._check_prepare()
                 # ... and recreate Thonny prompt
                 self.send_message(ToplevelResponse())
 
+    def _cmd_Run(self, cmd):
+        """Only for %run $EDITOR_CONTENT. start runs will be handled differently."""
+        if cmd.get("source"):
+            self._soft_reboot_without_running_main()
+            if self._submit_mode == PASTE_SUBMIT_MODE:
+                source = self._avoid_printing_expression_statements(cmd.source)
+            else:
+                source = cmd.source
+
+            self._execute(source, capture_output=False)
+            self._prepare_after_soft_reboot(False)
+        return {}
+
     def _cmd_execute_system_command(self, cmd):
         # Can't use stdin, because a thread is draining it
         returncode = execute_system_command(cmd, cwd=self._local_cwd, disconnect_stdin=True)
@@ -798,8 +1029,11 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
         self._check_sync_time()
         return super(BareMetalMicroPythonBackend, self)._cmd_write_file(cmd)
 
-    def _reset_environment(self):
-        self._soft_reboot_in_normal_mode_without_running_main()
+    def _cmd_prepare_disconnect(self, cmd):
+        logger.info("Preparing disconnect")
+        # NB! Don't let the mainloop see the prompt and act on it
+        self._connection.stop_reader()
+        self._write(NORMAL_MODE_CMD)
 
     def _delete_sorted_paths(self, paths):
         if not self._supports_directories():
@@ -849,7 +1083,7 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
     ) -> None:
         start_time = time.time()
 
-        if isinstance(self._connection, WebReplConnection):
+        if self._connected_over_webrepl():
             size = self._read_file_via_webrepl_file_protocol(source_path, target_fp, callback)
         else:
             # TODO: Is it better to read from mount when possible? Is the mount up to date when the file
@@ -869,7 +1103,7 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
         if hex_mode:
             self._execute_without_output("from binascii import hexlify as __temp_hexlify")
 
-        block_size = 1024
+        block_size = self._get_file_operation_block_size()
         file_size = self._get_file_size(source_path)
         num_bytes_read = 0
         while True:
@@ -909,7 +1143,7 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
         """
         Adapted from https://github.com/micropython/webrepl/blob/master/webrepl_cli.py
         """
-        assert isinstance(self._connection, WebReplConnection)
+        assert self._connected_over_webrepl()
 
         file_size = self._get_file_size(source_path)
 
@@ -949,7 +1183,7 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
     ) -> None:
         start_time = time.time()
 
-        if isinstance(self._connection, WebReplConnection):
+        if self._connected_over_webrepl():
             self._write_file_via_webrepl_file_protocol(source_fp, target_path, file_size, callback)
         else:
             try:
@@ -1056,7 +1290,8 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
             )
 
         bytes_sent = 0
-        block_size = 512
+        block_size = self._get_file_operation_block_size()
+
         while True:
             callback(bytes_sent, file_size)
             block = source_fp.read(block_size)
@@ -1119,7 +1354,7 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
         """
         Adapted from https://github.com/micropython/webrepl/blob/master/webrepl_cli.py
         """
-        assert isinstance(self._connection, WebReplConnection)
+        assert self._connected_over_webrepl()
 
         dest_fname = target_path.encode("utf-8")
         rec = struct.pack(
@@ -1269,7 +1504,7 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
                 return candidates[0]
 
     def _should_hexlify(self, path):
-        if "binascii" not in self._builtin_modules:
+        if "binascii" not in self._builtin_modules and "ubinascii" not in self._builtin_modules:
             return False
 
         for ext in (".py", ".txt", ".csv"):
@@ -1298,9 +1533,32 @@ class BareMetalMicroPythonBackend(MicroPythonBackend, UploadDownloadMixin):
     def _decode(self, data: bytes) -> str:
         return data.decode(ENCODING, errors="replace")
 
+    def _extract_block_without_splitting_chars(self, source_bytes: bytes) -> bytes:
+        i = self._write_block_size
+        while i > 1 and i < len(source_bytes) and is_continuation_byte(source_bytes[i]):
+            i -= 1
+
+        return source_bytes[:i]
+
+    def _get_file_operation_block_size(self):
+        # don't forget that the size may be expanded up to 4x where converted to Python
+        # bytes literal
+        if self._connected_to_microbit():
+            return 512
+        else:
+            return 1024
+
 
 def starts_with_continuation_byte(data):
-    return data and (data[0] & 0b11000000) == 0b10000000
+    return data and is_continuation_byte(data[0])
+
+
+def is_continuation_byte(byte):
+    return (byte & 0b11000000) == 0b10000000
+
+
+class RawPasteNotSupportedError(RuntimeError):
+    pass
 
 
 if __name__ == "__main__":
@@ -1332,9 +1590,13 @@ if __name__ == "__main__":
             # connection = DifficultSerialConnection(args["port"], BAUDRATE)
 
         if "circuitpython" in args.get("proxy_class", "").lower():
-            from thonny.plugins.circuitpython.cp_backend import CircuitPythonBackend
+            from thonny.plugins.circuitpython.cirpy_backend import CircuitPythonBackend
 
             backend = CircuitPythonBackend(connection, clean=args["clean"], args=args)
+        elif "pipico" in args.get("proxy_class", "").lower():
+            from thonny.plugins.rpi_pico.rpi_pico_backend import RaspberryPiPicoBackend
+
+            backend = RaspberryPiPicoBackend(connection, clean=args["clean"], args=args)
         else:
             backend = BareMetalMicroPythonBackend(connection, clean=args["clean"], args=args)
 
