@@ -11,8 +11,6 @@ import thonny
 from thonny import get_workbench, jedi_utils
 from thonny.codeview import get_syntax_options_for_tag
 
-python_tree = jedi_utils.import_python_tree()
-
 
 def create_bitmap_file(width, height, predicate, name):
 
@@ -127,6 +125,8 @@ def configure_text(text):
 
 
 def print_tree(node, level=0):
+    from parso.python import tree as python_tree
+
     indent = "  " * level
     # if (isinstance(node, python_tree.PythonNode) and node.type == "sim"
     if node.type in ("simple_stmt",) or isinstance(node, python_tree.Flow):
@@ -155,6 +155,7 @@ def add_tags(text):
 
     def tag_tree(node):
         nonlocal last_line, last_col
+        from parso.python import tree as python_tree
 
         if node.type == "simple_stmt" or isinstance(node, (python_tree.Flow, python_tree.Scope)):
 
