@@ -1,11 +1,12 @@
 # -*- coding: utf-8 -*-
-import platform
 import os
+import platform
 import sys
 
-from thonny.code import get_saved_current_script_filename
-from thonny import get_runner, terminal, get_workbench
-from thonny.common import get_exe_dirs, get_augmented_system_path
+from thonny import get_runner, get_workbench, terminal
+from thonny.common import get_augmented_system_path, get_exe_dirs
+from thonny.editors import get_saved_current_script_filename
+from thonny.languages import tr
 from thonny.running import get_environment_overrides_for_python_subprocess
 
 
@@ -17,6 +18,9 @@ def _open_system_shell():
     cwd = get_workbench().get_local_cwd()
 
     proxy = get_runner().get_backend_proxy()
+    if proxy and proxy.has_custom_system_shell():
+        proxy.open_custom_system_shell()
+        return
     if proxy and proxy.get_local_executable():
         target_executable = proxy.get_local_executable()
     else:
@@ -52,7 +56,7 @@ def load_plugin() -> None:
     get_workbench().add_command(
         "OpenSystemShell",
         "tools",
-        _("Open system shell..."),
+        tr("Open system shell..."),
         _open_system_shell,
         group=80,
         image="terminal",
