@@ -1,9 +1,9 @@
 # -*- coding: utf-8 -*-
-
-from tkinter import ttk, messagebox
+from tkinter import messagebox, ttk
 
 from thonny import get_runner, get_workbench
 from thonny.common import InlineCommand
+from thonny.languages import tr
 from thonny.memory import VariablesFrame
 
 
@@ -16,7 +16,7 @@ class VariablesView(VariablesFrame):
         self.back_button = ttk.Button(
             self.tree,
             style="Centered.TButton",
-            text=_("Back to\ncurrent frame"),
+            text=tr("Back to\ncurrent frame"),
             command=self._handle_back_button,
             width=15,
         )
@@ -33,7 +33,7 @@ class VariablesView(VariablesFrame):
     def _update_back_button(self, visible):
         if visible:
             assert self._last_active_info is not None
-            self.back_button.configure(text="Back to\n" + self._last_active_info[-1])
+            self.back_button.configure(text=tr("Back to\n%s") % self._last_active_info[-1])
             self.back_button.place(relx=1, x=-5, y=5, anchor="ne")
         else:
             self.back_button.place_forget()
@@ -52,10 +52,10 @@ class VariablesView(VariablesFrame):
     def _handle_get_globals_response(self, event):
         if "error" in event:
             self._clear_tree()
-            messagebox.showerror("Error querying global variables", event["error"])
+            messagebox.showerror("Error querying global variables", event["error"], master=self)
         elif "globals" not in event:
             self._clear_tree()
-            messagebox.showerror("Error querying global variables", str(event))
+            messagebox.showerror("Error querying global variables", str(event), master=self)
         else:
             self.show_globals(event["globals"], event["module_name"])
 
@@ -70,9 +70,9 @@ class VariablesView(VariablesFrame):
         # TODO: update only if something has changed
         self.update_variables(globals_)
         if module_name == "__main__":
-            self._set_tab_caption("Variables")
+            self._set_tab_caption(tr("Variables"))
         else:
-            self._set_tab_caption("Variables (%s)" % module_name)
+            self._set_tab_caption(tr("Variables") + " (%s)" % module_name)
 
         if is_active:
             self._last_active_info = (globals_, module_name)
@@ -135,4 +135,4 @@ class VariablesView(VariablesFrame):
 
 
 def load_plugin() -> None:
-    get_workbench().add_view(VariablesView, _("Variables"), "ne", default_position_key="AAA")
+    get_workbench().add_view(VariablesView, tr("Variables"), "ne", default_position_key="AAA")
