@@ -760,7 +760,7 @@ class BackendProxy:
         return None
 
     def interrupt(self):
-        """Tries to interrupt current command without reseting the backend"""
+        """Tries to interrupt current command without resetting the backend"""
         pass
 
     def destroy(self):
@@ -845,6 +845,9 @@ class SubprocessProxy(BackendProxy):
         else:
             self._executable = get_interpreter_for_subprocess()
 
+        if ".." in self._executable:
+            self._executable = os.path.normpath(self._executable)
+
         if not os.path.isfile(self._executable):
             raise UserError(
                 "Interpreter '%s' does not exist. Please check the configuration!"
@@ -853,6 +856,7 @@ class SubprocessProxy(BackendProxy):
         self._welcome_text = ""
 
         self._proc = None
+        self._terminated_readers = 0
         self._response_queue = None
         self._sys_path = []
         self._usersitepackages = None
@@ -864,11 +868,7 @@ class SubprocessProxy(BackendProxy):
     def _get_initial_cwd(self):
         return None
 
-    def _start_background_process(self, clean=None, extra_args=[]):
-        # deque, because in one occasion I need to put messages back
-        self._response_queue = collections.deque()
-
-        # prepare environment
+    def _get_environment(self):
         env = get_environment_for_python_subprocess(self._executable)
         # variables controlling communication with the back-end process
         env["PYTHONIOENCODING"] = "utf-8"
@@ -887,6 +887,11 @@ class SubprocessProxy(BackendProxy):
             env["THONNY_DEBUG"] = "1"
         elif "THONNY_DEBUG" in env:
             del env["THONNY_DEBUG"]
+        return env
+
+    def _start_background_process(self, clean=None, extra_args=[]):
+        # deque, because in one occasion I need to put messages back
+        self._response_queue = collections.deque()
 
         if not os.path.exists(self._executable):
             raise UserError(
@@ -922,13 +927,14 @@ class SubprocessProxy(BackendProxy):
             stdout=subprocess.PIPE,
             stderr=subprocess.PIPE,
             cwd=self._get_launch_cwd(),
-            env=env,
+            env=self._get_environment(),
             universal_newlines=True,
             creationflags=creationflags,
             **extra_params
         )
 
         # setup asynchronous output listeners
+        self._terminated_readers = 0
         Thread(target=self._listen_stdout, args=(self._proc.stdout,), daemon=True).start()
         Thread(target=self._listen_stderr, args=(self._proc.stderr,), daemon=True).start()
 
@@ -1005,7 +1011,7 @@ class SubprocessProxy(BackendProxy):
                 while len(message_queue) > 0:
                     sleep(0.005)
 
-        while self.process_is_alive():
+        while True:
             try:
                 data = stdout.readline()
             except IOError:
@@ -1043,9 +1049,11 @@ class SubprocessProxy(BackendProxy):
                                 )
                             )
 
+        self._terminated_readers += 1
+
     def _listen_stderr(self, stderr):
         # stderr is used only for debugger debugging
-        while self.process_is_alive():
+        while True:
             data = stderr.readline()
             if data == "":
                 break
@@ -1054,6 +1062,8 @@ class SubprocessProxy(BackendProxy):
                     BackendEvent("ProgramOutput", stream_name="stderr", data=data)
                 )
 
+        self._terminated_readers += 1
+
     def _store_state_info(self, msg):
         if "cwd" in msg:
             self._cwd = msg["cwd"]
@@ -1108,7 +1118,7 @@ class SubprocessProxy(BackendProxy):
 
     def fetch_next_message(self):
         if not self._response_queue or len(self._response_queue) == 0:
-            if self.is_terminated():
+            if self.is_terminated() and self._terminated_readers == 2:
                 raise BackendTerminatedError(self._proc.returncode if self._proc else None)
             else:
                 return None
