@@ -35,12 +35,23 @@ from thonny.ui_utils import (
     ems_to_pixels,
 )
 
+OBJECT_INFO_START_REGEX_STR = re.escape(OBJECT_LINK_START).replace("%d", r"-?\d+")
+OBJECT_INFO_START_REGEX = re.compile(OBJECT_INFO_START_REGEX_STR)
+OBJECT_INFO_END_REGEX_STR = re.escape(OBJECT_LINK_END)
+OBJECT_INFO_END_REGEX = re.compile(OBJECT_INFO_END_REGEX_STR)
+
 logger = logging.getLogger(__name__)
 
 _CLEAR_SHELL_DEFAULT_SEQ = select_sequence("<Control-l>", "<Command-k>")
 
 # NB! Don't add parens without refactoring split procedure!
-OUTPUT_SPLIT_REGEX = re.compile(r"(\x1B\[[0-?]*[ -/]*[@-~]|[\a\b\r])")
+
+TERMINAL_CONTROL_REGEX_STR = r"\x1B\[[0-?]*[ -/]*[@-~]|[\a\b\r]"
+TERMINAL_CONTROL_REGEX = re.compile(TERMINAL_CONTROL_REGEX_STR)
+OUTPUT_SPLIT_REGEX = re.compile(
+    "(%s|%s|%s)"
+    % (TERMINAL_CONTROL_REGEX_STR, OBJECT_INFO_START_REGEX_STR, OBJECT_INFO_END_REGEX_STR)
+)
 NUMBER_SPLIT_REGEX = re.compile(r"((?<!\w)[-+]?[0-9]*\.?[0-9]+\b)")
 SIMPLE_URL_SPLIT_REGEX = re.compile(
     r"(https?:\/\/[\w\/.:\-\?#=%&]+[\w\/]|data:image\/[a-z]+;base64,[A-Za-z0-9\/=\+]+)"
@@ -192,6 +203,7 @@ class ShellView(tk.PanedWindow):
 
     def print_error(self, txt):
         self.text._insert_text_directly(txt, ("io", "stderr"))
+        self.text.see("end")
 
     def insert_command_link(self, txt, handler):
         self.text._insert_command_link(txt, handler)
@@ -372,7 +384,7 @@ class BaseShellText(EnhancedTextWithLogging, SyntaxText):
         self.mark_set("command_io_start", "1.0")
         self.mark_gravity("command_io_start", "left")
 
-        self.active_object_tags = set()
+        self.active_extra_tags = []
 
         self.update_tabs()
 
@@ -419,7 +431,6 @@ class BaseShellText(EnhancedTextWithLogging, SyntaxText):
             self.tag_add("before_io", "output_insert -1 line linestart")
 
         self._update_visible_io(None)
-        self._render_object_links("end")
 
     def _handle_toplevel_response(self, msg: ToplevelResponse) -> None:
         if msg.get("error"):
@@ -438,33 +449,6 @@ class BaseShellText(EnhancedTextWithLogging, SyntaxText):
             self._insert_text_directly(welcome_text, ("welcome",))
             self.see("end")
 
-        if "value_info" in msg:
-            num_stripped_question_marks = getattr(msg, "num_stripped_question_marks", 0)
-            if num_stripped_question_marks > 0:
-                # show the value in object inspector
-                get_workbench().event_generate("ObjectSelect", object_id=msg["value_info"].id)
-            else:
-                # show the value in shell
-                value_repr = shorten_repr(msg["value_info"].repr, 10000)
-                if value_repr != "None":
-                    if get_workbench().in_heap_mode():
-                        value_repr = memory.format_object_id(msg["value_info"].id)
-                    object_tag = "object_" + str(msg["value_info"].id)
-                    self._insert_text_directly(value_repr + "\n", ("toplevel", "value", object_tag))
-                    if running_on_mac_os():
-                        sequence = "<Command-Button-1>"
-                    else:
-                        sequence = "<Control-Button-1>"
-                    self.tag_bind(
-                        object_tag,
-                        sequence,
-                        lambda _: get_workbench().event_generate(
-                            "ObjectSelect", object_id=msg["value_info"].id
-                        ),
-                    )
-
-                    self.active_object_tags.add(object_tag)
-
         self.mark_set("output_end", self.index("end-1c"))
         self._discard_old_content()
         self._update_visible_io(None)
@@ -489,18 +473,15 @@ class BaseShellText(EnhancedTextWithLogging, SyntaxText):
         return get_workbench().get_option("shell.squeeze_threshold")
 
     def _append_to_io_queue(self, data, stream_name):
-        if self.tty_mode:
-            # Make sure ANSI CSI codes are stored as separate events
-            # TODO: try to complete previously submitted incomplete code
-            parts = re.split(OUTPUT_SPLIT_REGEX, data)
-            for part in parts:
-                if part:  # split may produce empty string in the beginning or start
-                    # split the data so that very long lines separated
-                    for block in re.split("(.{%d,})" % (self._get_squeeze_threshold() + 1), part):
-                        if block:
-                            self._queued_io_events.append((block, stream_name))
-        else:
-            self._queued_io_events.append((data, stream_name))
+        # Make sure ANSI CSI codes and object links are stored as separate events
+        # TODO: try to complete previously submitted incomplete code
+        parts = re.split(OUTPUT_SPLIT_REGEX, data)
+        for part in parts:
+            if part:  # split may produce empty string in the beginning or start
+                # split the data so that very long lines separated
+                for block in re.split("(.{%d,})" % (self._get_squeeze_threshold() + 1), part):
+                    if block:
+                        self._queued_io_events.append((block, stream_name))
 
     def _update_visible_io(self, target_num_visible_chars):
         current_num_visible_chars = sum(map(lambda x: len(x[0]), self._applied_io_events))
@@ -534,13 +515,13 @@ class BaseShellText(EnhancedTextWithLogging, SyntaxText):
         self.mark_set("output_end", self.index("end-1c"))
         self.see("end")
 
-    def _apply_io_event(self, data, stream_name, extra_tags=set()):
+    def _apply_io_event(self, data, stream_name):
         if not data:
             return
 
         original_data = data
 
-        if self.tty_mode and re.match(OUTPUT_SPLIT_REGEX, data):
+        if self.tty_mode and re.match(TERMINAL_CONTROL_REGEX, data):
             if data == "\a":
                 get_workbench().bell()
             elif data == "\b":
@@ -555,9 +536,34 @@ class BaseShellText(EnhancedTextWithLogging, SyntaxText):
                 # It makes sense, but Ubuntu terminal does not confirm it.
                 # For now I'm just trimming stderr color codes
                 self._update_ansi_attributes(data)
+            else:
+                logger.warning("Don't know what to do with %r" % data)
 
+        elif re.match(OBJECT_INFO_START_REGEX, data):
+            id_str = data[data.index("=") + 1 : data.index("]")]
+            self.active_extra_tags.append("value")
+            self.active_extra_tags.append(id_str)
+            if get_workbench().get_option("shell.auto_inspect_values"):
+                get_workbench().event_generate("ObjectSelect", object_id=int(id_str))
+
+            if get_workbench().in_heap_mode():
+                self._insert_text_directly(
+                    memory.format_object_id(int(id_str)), tuple(self.active_extra_tags)
+                )
+
+        elif re.match(OBJECT_INFO_END_REGEX, data):
+            self.active_extra_tags.pop()
+            self.active_extra_tags.pop()
+
+        elif "value" in self.active_extra_tags and get_workbench().in_heap_mode():
+            # id was already printed and value should be suppressed
+            pass
         else:
-            tags = extra_tags | {"io", stream_name}
+            if "value" in self.active_extra_tags:
+                tags = set(self.active_extra_tags)
+            else:
+                tags = set(self.active_extra_tags) | {"io", stream_name}
+
             if stream_name == "stdout" and self.tty_mode:
                 tags |= self._get_ansi_tags()
 
@@ -571,7 +577,7 @@ class BaseShellText(EnhancedTextWithLogging, SyntaxText):
                 and not (data.startswith(OBJECT_LINK_START))
             ):
                 self._io_cursor_offset = 0  # ignore the effect of preceding \r and \b
-                actual_text = self._remove_object_link_markers(data)
+                actual_text = data
                 button_text = actual_text[:70] + " …"
                 btn = tk.Label(
                     self,
@@ -636,62 +642,6 @@ class BaseShellText(EnhancedTextWithLogging, SyntaxText):
 
         self._applied_io_events.append((original_data, stream_name))
 
-    def _remove_object_link_markers(self, data):
-        end_pos = data.find(OBJECT_LINK_END)
-        if end_pos < 0:
-            return data
-
-        at_pos = data.rfind("@", 0, end_pos)
-        if at_pos < 0:
-            return data
-
-        data = data[:at_pos] + data[end_pos + len(OBJECT_LINK_END) :]
-        return data.replace(OBJECT_LINK_START, "")
-
-    def _render_object_links(self, start):
-        closer_start = self.search(
-            OBJECT_LINK_END, index=start, stopindex="command_io_start", backwards=True, elide=False
-        )
-        if not closer_start:
-            return
-
-        id_start = self.search(
-            "@", index=closer_start, stopindex="command_io_start", backwards=True
-        )
-        if not id_start:
-            logger.warning("Can't find object id")
-            return
-
-        id_str = self.get(id_start, closer_start)[1:]
-        try:
-            int(id_str)
-        except ValueError:
-            logger.warning("Can't parse object id: " + id_str)
-            return
-
-        self.direct_delete(id_start, "%s + %d chars" % (closer_start, len(OBJECT_LINK_END)))
-
-        opener_start = self.search(
-            OBJECT_LINK_START, index=id_start, stopindex="command_io_start", backwards=True
-        )
-        value_start = "%s + %d chars" % (opener_start, len(OBJECT_LINK_START))
-
-        self.tag_remove("io", value_start, closer_start)
-        self.tag_add("value", value_start, closer_start)
-        self.tag_add(id_str, value_start, closer_start)
-
-        if get_workbench().in_heap_mode():
-            id_repr = memory.format_object_id(int(id_str))
-            self.direct_delete(value_start, id_start)
-            self.direct_insert(value_start, id_repr, ("value", id_str))
-
-        self.direct_delete(opener_start, value_start)
-
-        self._render_object_links(opener_start)
-
-        if get_workbench().get_option("shell.auto_inspect_values"):
-            get_workbench().event_generate("ObjectSelect", object_id=int(id_str))
-
     def _show_squeezed_text(self, button):
         dlg = SqueezedTextDialog(self, button)
         show_dialog(dlg)
@@ -1383,7 +1333,7 @@ class BaseShellText(EnhancedTextWithLogging, SyntaxText):
         pos = "@%d,%d" % (event.x, event.y)
         tags = self.tag_names(pos)
         for tag in tags:
-            if tag.isnumeric():
+            if tag.isnumeric() or tag[0] == "-" and tag[1:].isnumeric():
                 get_workbench().show_view("ObjectInspector", set_focus=False)
                 get_workbench().update_idletasks()
                 get_workbench().event_generate("ObjectSelect", object_id=int(tag))
@@ -1395,9 +1345,9 @@ class BaseShellText(EnhancedTextWithLogging, SyntaxText):
             return
 
         # select whole value unless user has started a partial selection
-        if not self.tag_nextrange("sel", rng[0], rng[1]):
-            self.tag_remove("sel", "1.0", "end")
-            self.tag_add("sel", rng[0], rng[1])
+        # if not self.tag_nextrange("sel", rng[0], rng[1]):
+        #    self.tag_remove("sel", "1.0", "end")
+        #    self.tag_add("sel", rng[0], rng[1])
 
     def _handle_hyperlink(self, event):
         import webbrowser
@@ -1429,8 +1379,8 @@ class BaseShellText(EnhancedTextWithLogging, SyntaxText):
                     if SIMPLE_URL_SPLIT_REGEX.match(url):
                         webbrowser.open(url)
 
-        except Exception:
-            traceback.print_exc()
+        except Exception as e:
+            logger.exception("Could not handle hyperlink click", exc_info=e)
 
     def _show_user_exception(self, user_exception):
         for line, frame_id, *_ in user_exception["items"]:
@@ -1465,16 +1415,18 @@ class BaseShellText(EnhancedTextWithLogging, SyntaxText):
     def _clear_content(self, cut_idx):
         proposed_cut_float = float(cut_idx)
         for btn in list(self._squeeze_buttons):
-            idx = self.index(btn)
-            if idx is None or float(idx) < proposed_cut_float:
-                self._squeeze_buttons.remove(btn)
-                # looks like the widgets are not fully GC-d.
-                # At least avoid leaking big chunks of texts
-                try:
+            try:
+                idx = self.index(btn)
+                if idx is None or idx == "" or float(idx) < proposed_cut_float:
+                    self._squeeze_buttons.remove(btn)
+                    # looks like the widgets are not fully GC-d.
+                    # At least avoid leaking big chunks of texts
                     btn.contained_text = None
                     btn.destroy()
-                except:
-                    logger.warning("Could not destroy a squeeze button")
+            except Exception as e:
+                logger.warning("Problem with a squeeze button, removing it", exc_info=e)
+                if btn in self._squeeze_buttons:
+                    self._squeeze_buttons.remove(btn)
 
         self.direct_delete("0.1", cut_idx)
 
@@ -1496,8 +1448,8 @@ class BaseShellText(EnhancedTextWithLogging, SyntaxText):
         self.tag_add("inactive", "1.0", end_index)
         self.tag_remove("value", "1.0", end_index)
 
-        while len(self.active_object_tags) > 0:
-            self.tag_remove(self.active_object_tags.pop(), "1.0", "end")
+        while len(self.active_extra_tags) > 0:
+            self.tag_remove(self.active_extra_tags.pop(), "1.0", "end")
 
     def get_lines_above_viewport_bottom(self, tag_name, n):
         end_index = self.index("@%d,%d lineend" % (self.winfo_height(), self.winfo_height()))
@@ -1859,7 +1811,7 @@ class PlotterCanvas(tk.Canvas):
                     y,
                     text=marker,
                     anchor="sw",
-                    fill=self.colors[i - 1 % len(self.colors)],
+                    fill=self.colors[(i - 1) % len(self.colors)],
                     tags=("legend",),
                 )
                 x += marker_width
