@@ -3,35 +3,38 @@ import collections
 import logging
 import os
 import platform
+import queue
 import re
-import shutil
 import signal
 import subprocess
+import sys
 import textwrap
 import threading
 import time
 import tkinter as tk
 import tkinter.font
 import traceback
+import warnings
 from tkinter import filedialog, messagebox, ttk
 from typing import Callable, List, Optional, Tuple, Union  # @UnusedImport
 
+from _tkinter import TclError
+
 from thonny import get_workbench, misc_utils, tktextext
 from thonny.common import TextRange
+from thonny.languages import get_button_padding, tr
 from thonny.misc_utils import (
     running_on_linux,
     running_on_mac_os,
-    running_on_windows,
     running_on_rpi,
+    running_on_windows,
 )
 from thonny.tktextext import TweakableText
-import sys
-from _tkinter import TclError
-import webbrowser
-from thonny.languages import get_button_padding
 
 PARENS_REGEX = re.compile(r"[\(\)\{\}\[\]]")
 
+logger = logging.getLogger(__name__)
+
 
 class CommonDialog(tk.Toplevel):
     def __init__(self, master=None, cnf={}, **kw):
@@ -45,6 +48,12 @@ class CommonDialog(tk.Toplevel):
             if focussed_widget:
                 focussed_widget.focus_set()
 
+    def get_padding(self):
+        return ems_to_pixels(2)
+
+    def get_internal_padding(self):
+        return self.get_padding() // 4
+
 
 class CommonDialogEx(CommonDialog):
     def __init__(self, master=None, cnf={}, **kw):
@@ -63,6 +72,79 @@ class CommonDialogEx(CommonDialog):
         self.destroy()
 
 
+class QueryDialog(CommonDialogEx):
+    def __init__(
+        self,
+        master,
+        title: str,
+        prompt: str,
+        initial_value: str = "",
+        options: List[str] = [],
+        entry_width: Optional[int] = None,
+    ):
+        super().__init__(master)
+        self.var = tk.StringVar(value=initial_value)
+        self.result = None
+
+        margin = self.get_padding()
+        spacing = margin // 2
+
+        self.title(title)
+        self.prompt_label = ttk.Label(self.main_frame, text=prompt)
+        self.prompt_label.grid(row=1, column=1, columnspan=2, padx=margin, pady=(margin, spacing))
+
+        if options:
+            self.entry_widget = ttk.Combobox(
+                self.main_frame, textvariable=self.var, values=options, height=15, width=entry_width
+            )
+        else:
+            self.entry_widget = ttk.Entry(self.main_frame, textvariable=self.var, width=entry_width)
+
+        self.entry_widget.bind("<Return>", self.on_ok, True)
+        self.entry_widget.bind("<KP_Enter>", self.on_ok, True)
+
+        self.entry_widget.grid(
+            row=3, column=1, columnspan=2, sticky="we", padx=margin, pady=(0, margin)
+        )
+
+        self.ok_button = ttk.Button(
+            self.main_frame, text=tr("OK"), command=self.on_ok, default="active"
+        )
+        self.ok_button.grid(row=5, column=1, padx=(margin, spacing), pady=(0, margin), sticky="e")
+        self.cancel_button = ttk.Button(self.main_frame, text=tr("Cancel"), command=self.on_cancel)
+        self.cancel_button.grid(row=5, column=2, padx=(0, margin), pady=(0, margin), sticky="e")
+
+        self.main_frame.columnconfigure(1, weight=1)
+
+        self.entry_widget.focus_set()
+
+    def on_ok(self, event=None):
+        self.result = self.var.get()
+        self.destroy()
+
+    def on_cancel(self, event=None):
+        self.result = None
+        self.destroy()
+
+    def get_result(self) -> Optional[str]:
+        return self.result
+
+
+def ask_string(
+    title: str,
+    prompt: str,
+    initial_value: str = "",
+    options: List[str] = [],
+    entry_width: Optional[int] = None,
+    master=None,
+):
+    dlg = QueryDialog(
+        master, title, prompt, initial_value=initial_value, options=options, entry_width=entry_width
+    )
+    show_dialog(dlg, master)
+    return dlg.get_result()
+
+
 class CustomMenubar(ttk.Frame):
     def __init__(self, master):
         ttk.Frame.__init__(self, master, style="CustomMenubar.TFrame")
@@ -138,8 +220,7 @@ class AutomaticPanedWindow(tk.PanedWindow):
     """
 
     def __init__(self, master, position_key=None, preferred_size_in_pw=None, **kwargs):
-
-        tk.PanedWindow.__init__(self, master, **kwargs)
+        tk.PanedWindow.__init__(self, master, border=0, **kwargs)
         self._pane_minsize = 100
         self.position_key = position_key
         self._restoring_pane_sizes = False
@@ -340,7 +421,7 @@ class AutomaticPanedWindow(tk.PanedWindow):
             self.master.insert("auto", self)
 
     def _update_appearance(self, event=None):
-        self.configure(sashwidth=lookup_style_option("Sash", "sashthickness", 10))
+        self.configure(sashwidth=lookup_style_option("Sash", "sashthickness", ems_to_pixels(0.6)))
         self.configure(background=lookup_style_option("TPanedWindow", "background"))
 
 
@@ -364,9 +445,9 @@ class ClosableNotebook(ttk.Notebook):
 
     def create_tab_menu(self):
         menu = tk.Menu(self.winfo_toplevel(), tearoff=False, **get_style_configuration("Menu"))
-        menu.add_command(label=_("Close"), command=self._close_tab_from_menu)
-        menu.add_command(label=_("Close others"), command=self._close_other_tabs)
-        menu.add_command(label=_("Close all"), command=self.close_tabs)
+        menu.add_command(label=tr("Close"), command=self._close_tab_from_menu)
+        menu.add_command(label=tr("Close others"), command=self._close_other_tabs)
+        menu.add_command(label=tr("Close all"), command=self.close_tabs)
         return menu
 
     def _letf_btn_press(self, event):
@@ -608,6 +689,7 @@ class TreeFrame(ttk.Frame):
         columns,
         displaycolumns="#all",
         show_scrollbar=True,
+        show_statusbar=False,
         borderwidth=0,
         relief="flat",
         **tree_kw
@@ -618,7 +700,9 @@ class TreeFrame(ttk.Frame):
             self, orient=tk.VERTICAL, style=scrollbar_style("Vertical")
         )
         if show_scrollbar:
-            self.vert_scrollbar.grid(row=0, column=1, sticky=tk.NSEW)
+            self.vert_scrollbar.grid(
+                row=0, column=1, sticky=tk.NSEW, rowspan=2 if show_statusbar else 1
+            )
 
         self.tree = ttk.Treeview(
             self,
@@ -635,6 +719,12 @@ class TreeFrame(ttk.Frame):
         self.tree.bind("<<TreeviewSelect>>", self.on_select, "+")
         self.tree.bind("<Double-Button-1>", self.on_double_click, "+")
 
+        if show_statusbar:
+            self.statusbar = ttk.Frame(self)
+            self.statusbar.grid(row=1, column=0, sticky="nswe")
+        else:
+            self.statusbar = None
+
     def _clear_tree(self):
         for child_id in self.tree.get_children():
             self.tree.delete(child_id)
@@ -714,54 +804,38 @@ def set_zoomed(toplevel, value):
 
 
 class EnhancedTextWithLogging(tktextext.EnhancedText):
-    def __init__(
-        self,
-        master=None,
-        style="Text",
-        tag_current_line=False,
-        indent_with_tabs=False,
-        replace_tabs=False,
-        cnf={},
-        **kw
-    ):
+    def __init__(self, master=None, style="Text", tag_current_line=False, cnf={}, **kw):
         super().__init__(
-            master=master,
-            style=style,
-            tag_current_line=tag_current_line,
-            indent_with_tabs=indent_with_tabs,
-            replace_tabs=replace_tabs,
-            cnf=cnf,
-            **kw
+            master=master, style=style, tag_current_line=tag_current_line, cnf=cnf, **kw
         )
 
         self._last_event_changed_line_count = False
 
     def direct_insert(self, index, chars, tags=None, **kw):
-        try:
-            # try removing line numbers
-            # TODO: shouldn't it take place only on paste?
-            # TODO: does it occur when opening a file with line numbers in it?
-            # if self._propose_remove_line_numbers and isinstance(chars, str):
-            #    chars = try_remove_linenumbers(chars, self)
+        # try removing line numbers
+        # TODO: shouldn't it take place only on paste?
+        # TODO: does it occur when opening a file with line numbers in it?
+        # if self._propose_remove_line_numbers and isinstance(chars, str):
+        #    chars = try_remove_linenumbers(chars, self)
 
-            concrete_index = self.index(index)
-            line_before = self.get(concrete_index + " linestart", concrete_index + " lineend")
-            self._last_event_changed_line_count = "\n" in chars
-            return tktextext.EnhancedText.direct_insert(self, index, chars, tags=tags, **kw)
-        finally:
-            line_after = self.get(concrete_index + " linestart", concrete_index + " lineend")
-            trivial_for_coloring, trivial_for_parens = self._is_trivial_edit(
-                chars, line_before, line_after
-            )
-            get_workbench().event_generate(
-                "TextInsert",
-                index=concrete_index,
-                text=chars,
-                tags=tags,
-                text_widget=self,
-                trivial_for_coloring=trivial_for_coloring,
-                trivial_for_parens=trivial_for_parens,
-            )
+        concrete_index = self.index(index)
+        line_before = self.get(concrete_index + " linestart", concrete_index + " lineend")
+        self._last_event_changed_line_count = "\n" in chars
+        result = tktextext.EnhancedText.direct_insert(self, index, chars, tags=tags, **kw)
+        line_after = self.get(concrete_index + " linestart", concrete_index + " lineend")
+        trivial_for_coloring, trivial_for_parens = self._is_trivial_edit(
+            chars, line_before, line_after
+        )
+        get_workbench().event_generate(
+            "TextInsert",
+            index=concrete_index,
+            text=chars,
+            tags=tags,
+            text_widget=self,
+            trivial_for_coloring=trivial_for_coloring,
+            trivial_for_parens=trivial_for_parens,
+        )
+        return result
 
     def direct_delete(self, index1, index2=None, **kw):
         try:
@@ -1027,6 +1101,8 @@ class ToolTip:
                 tw.tk.call(
                     "::tk::unsupported::MacWindowStyle", "style", tw._w, "help", "noActivates"
                 )
+                if get_tk_version_info() >= (8, 6, 10) and running_on_mac_os():
+                    tw.wm_overrideredirect(1)
             except tk.TclError:
                 pass
         else:
@@ -1054,6 +1130,7 @@ class ToolTip:
 def create_tooltip(widget, text, **kw):
     options = get_style_configuration("Tooltip").copy()
     options.setdefault("background", "#ffffe0")
+    options.setdefault("foreground", "#000000")
     options.setdefault("relief", "solid")
     options.setdefault("borderwidth", 1)
     options.setdefault("padx", 1)
@@ -1203,7 +1280,6 @@ class NoteBox(CommonDialog):
         lines = self._current_chars.splitlines()
         max_line_width = 0
         for line in lines:
-
             max_line_width = max(max_line_width, font.measure(line))
 
         width = min(max_line_width, self._max_default_width) + self.padx * 2 + 2
@@ -1246,45 +1322,66 @@ def get_widget_offset_from_toplevel(widget):
     return x, y
 
 
-def create_string_var(value, modification_listener=None):
-    """Creates a tk.StringVar with "modified" attribute
-    showing whether the variable has been modified after creation"""
-    return _create_var(tk.StringVar, value, modification_listener)
+class EnhancedVar(tk.Variable):
+    def __init__(self, master=None, value=None, name=None, modification_listener=None):
+        if master is not None and not isinstance(master, (tk.Widget, tk.Wm)):
+            raise TypeError("First positional argument 'master' must be None, Widget or Wm")
 
+        super().__init__(master=master, value=value, name=name)
+        self.modified = False
+        self.modification_listener = modification_listener
+        if sys.version_info < (3, 6):
+            self.trace("w", self._on_write)
+        else:
+            self.trace_add("write", self._on_write)
 
-def create_int_var(value, modification_listener=None):
-    """See create_string_var"""
-    return _create_var(tk.IntVar, value, modification_listener)
-
-
-def create_double_var(value, modification_listener=None):
-    """See create_string_var"""
-    return _create_var(tk.DoubleVar, value, modification_listener)
-
-
-def create_boolean_var(value, modification_listener=None):
-    """See create_string_var"""
-    return _create_var(tk.BooleanVar, value, modification_listener)
-
-
-def _create_var(class_, value, modification_listener):
-    var = class_(value=value)
-    var.modified = False
-
-    def on_write(*args):
-        var.modified = True
-        if modification_listener:
+    def _on_write(self, *args):
+        self.modified = True
+        if self.modification_listener:
             try:
-                modification_listener()
+                self.modification_listener()
             except Exception:
                 # Otherwise whole process will be brought down
                 # because for some reason Tk tries to call non-existing method
                 # on variable
                 get_workbench().report_exception()
 
-    # TODO: https://bugs.python.org/issue22115 (deprecation warning)
-    var.trace("w", on_write)
-    return var
+
+class EnhancedStringVar(EnhancedVar, tk.StringVar):
+    pass
+
+
+class EnhancedIntVar(EnhancedVar, tk.IntVar):
+    pass
+
+
+class EnhancedBooleanVar(EnhancedVar, tk.BooleanVar):
+    pass
+
+
+class EnhancedDoubleVar(EnhancedVar, tk.DoubleVar):
+    pass
+
+
+def create_string_var(value, modification_listener=None) -> EnhancedStringVar:
+    """Creates a tk.StringVar with "modified" attribute
+    showing whether the variable has been modified after creation"""
+    return EnhancedStringVar(None, value, None, modification_listener)
+
+
+def create_int_var(value, modification_listener=None) -> EnhancedIntVar:
+    """See create_string_var"""
+    return EnhancedIntVar(None, value, None, modification_listener)
+
+
+def create_double_var(value, modification_listener=None) -> EnhancedDoubleVar:
+    """See create_string_var"""
+    return EnhancedDoubleVar(None, value, None, modification_listener)
+
+
+def create_boolean_var(value, modification_listener=None) -> EnhancedBooleanVar:
+    """See create_string_var"""
+    return EnhancedBooleanVar(None, value, None, modification_listener)
 
 
 def shift_is_pressed(event_state):
@@ -1293,6 +1390,12 @@ def shift_is_pressed(event_state):
     return event_state & 0x0001
 
 
+def caps_lock_is_on(event_state):
+    # http://infohost.nmt.edu/tcc/help/pubs/tkinter/web/event-handlers.html
+    # http://stackoverflow.com/q/32426250/261181
+    return event_state & 0x0002
+
+
 def control_is_pressed(event_state):
     # http://infohost.nmt.edu/tcc/help/pubs/tkinter/web/event-handlers.html
     # http://stackoverflow.com/q/32426250/261181
@@ -1358,6 +1461,7 @@ def try_remove_linenumbers(text, master):
             title="Remove linenumbers",
             message="Do you want to remove linenumbers from pasted text?",
             default=messagebox.YES,
+            master=master,
         ):
             return remove_line_numbers(text)
         else:
@@ -1399,7 +1503,6 @@ def center_window(win, master=None):
 
 
 def assign_geometry(win, master=None, min_left=0, min_top=0):
-
     if master is None:
         master = tk._default_root
 
@@ -1505,10 +1608,10 @@ class FileCopyDialog(CommonDialog):
         self.rowconfigure(0, weight=1)
         self.columnconfigure(0, weight=1)
 
-        self.title(_("Copying"))
+        self.title(tr("Copying"))
 
         if description is None:
-            description = _("Copying\n  %s\nto\n  %s") % (source, destination)
+            description = tr("Copying\n  %s\nto\n  %s") % (source, destination)
 
         label = ttk.Label(main_frame, text=description)
         label.grid(row=0, column=0, columnspan=2, sticky="nw", padx=15, pady=15)
@@ -1516,7 +1619,7 @@ class FileCopyDialog(CommonDialog):
         self._bar = ttk.Progressbar(main_frame, maximum=os.path.getsize(source), length=200)
         self._bar.grid(row=1, column=0, columnspan=2, sticky="nsew", padx=15, pady=0)
 
-        self._cancel_button = ttk.Button(main_frame, text=_("Cancel"), command=self._cancel)
+        self._cancel_button = ttk.Button(main_frame, text=tr("Cancel"), command=self._cancel)
         self._cancel_button.grid(row=2, column=1, sticky="ne", padx=15, pady=15)
         self._bar.focus_set()
 
@@ -1569,41 +1672,45 @@ class FileCopyDialog(CommonDialog):
         self._close()
 
 
-class ChoiceDialog(CommonDialog):
+class ChoiceDialog(CommonDialogEx):
     def __init__(
-        self, master=None, title="Choose one", question: str = "Choose one:", choices=[]
+        self,
+        master=None,
+        title="Choose one",
+        question: str = "Choose one:",
+        choices=[],
+        initial_choice_index=None,
     ) -> None:
         super().__init__(master=master)
 
         self.title(title)
         self.resizable(False, False)
 
-        self.columnconfigure(0, weight=1)
+        self.main_frame.columnconfigure(0, weight=1)
 
         row = 0
-        question_label = ttk.Label(self, text=question)
+        question_label = ttk.Label(self.main_frame, text=question)
         question_label.grid(row=row, column=0, columnspan=2, sticky="w", padx=20, pady=20)
         row += 1
 
-        self.var = tk.StringVar()
+        self.var = tk.StringVar("")
+        if initial_choice_index is not None:
+            self.var.set(choices[initial_choice_index])
         for choice in choices:
-            rb = ttk.Radiobutton(self, text=choice, variable=self.var, value=choice)
+            rb = ttk.Radiobutton(self.main_frame, text=choice, variable=self.var, value=choice)
             rb.grid(row=row, column=0, columnspan=2, sticky="w", padx=20)
             row += 1
 
-        ok_button = ttk.Button(self, text="OK", command=self._ok, default="active")
+        ok_button = ttk.Button(self.main_frame, text=tr("OK"), command=self._ok, default="active")
         ok_button.grid(row=row, column=0, sticky="e", pady=20)
 
-        cancel_button = ttk.Button(self, text="Cancel", command=self._cancel)
+        cancel_button = ttk.Button(self.main_frame, text=tr("Cancel"), command=self._cancel)
         cancel_button.grid(row=row, column=1, sticky="e", padx=20, pady=20)
 
         self.bind("<Escape>", self._cancel, True)
         self.bind("<Return>", self._ok, True)
         self.protocol("WM_DELETE_WINDOW", self._cancel)
 
-        if misc_utils.running_on_mac_os():
-            self.configure(background="systemSheetBackground")
-
     def _ok(self):
         self.result = self.var.get()
         if not self.result:
@@ -1645,12 +1752,15 @@ class LongTextDialog(CommonDialog):
         self._text.text.see("1.0")
 
         copy_button = ttk.Button(
-            main_frame, command=self._copy, text=_("Copy to clipboard"), width=20
+            main_frame, command=self._copy, text=tr("Copy to clipboard"), width=20
         )
         copy_button.grid(row=2, column=0, sticky="w", padx=20, pady=(0, 20))
 
-        close_button = ttk.Button(main_frame, command=self._close, text=_("Close"))
+        close_button = ttk.Button(
+            main_frame, command=self._close, text=tr("Close"), default="active"
+        )
         close_button.grid(row=2, column=1, sticky="w", padx=20, pady=(0, 20))
+        close_button.focus_set()
 
         main_frame.columnconfigure(0, weight=1)
         main_frame.rowconfigure(1, weight=1)
@@ -1667,173 +1777,17 @@ class LongTextDialog(CommonDialog):
 
 
 def ask_one_from_choices(
-    master=None, title="Choose one", question: str = "Choose one:", choices=[]
+    master=None,
+    title="Choose one",
+    question: str = "Choose one:",
+    choices=[],
+    initial_choice_index=None,
 ):
-    dlg = ChoiceDialog(master, title, question, choices)
+    dlg = ChoiceDialog(master, title, question, choices, initial_choice_index)
     show_dialog(dlg, master)
     return dlg.result
 
 
-class SubprocessDialog(CommonDialog):
-    """Shows incrementally the output of given subprocess.
-    Allows cancelling"""
-
-    def __init__(
-        self, master, proc, title, long_description=None, autoclose=True, conclusion="Done."
-    ):
-        self._closed = False
-        self._proc = proc
-        self.stdout = ""
-        self.stderr = ""
-        self._stdout_thread = None
-        self._stderr_thread = None
-        self.returncode = None
-        self.cancelled = False
-        self._autoclose = autoclose
-        self._event_queue = collections.deque()
-        self._conclusion = conclusion
-
-        super().__init__(master)
-
-        self.rowconfigure(0, weight=1)
-        self.columnconfigure(0, weight=1)
-        main_frame = ttk.Frame(self)  # To get styled background
-        main_frame.grid(sticky="nsew")
-
-        text_font = tk.font.nametofont("TkFixedFont").copy()
-        text_font["size"] = int(text_font["size"] * 0.9)
-        text_font["family"] = "Courier" if running_on_mac_os() else "Courier New"
-        text_frame = tktextext.TextFrame(
-            main_frame,
-            read_only=True,
-            horizontal_scrollbar=False,
-            background=lookup_style_option("TFrame", "background"),
-            font=text_font,
-            wrap="word",
-        )
-        text_frame.grid(row=0, column=0, sticky=tk.NSEW, padx=15, pady=15)
-        self.text = text_frame.text
-        self.text["width"] = 60
-        self.text["height"] = 7
-        if long_description is not None:
-            self.text.direct_insert("1.0", long_description + "\n\n")
-
-        self.button = ttk.Button(main_frame, text=_("Cancel"), command=self._close)
-        self.button.grid(row=1, column=0, pady=(0, 15))
-
-        main_frame.rowconfigure(0, weight=1)
-        main_frame.columnconfigure(0, weight=1)
-
-        self.title(title)
-        if misc_utils.running_on_mac_os():
-            self.configure(background="systemSheetBackground")
-        # self.resizable(height=tk.FALSE, width=tk.FALSE)
-        self.text.focus_set()
-
-        self.bind(
-            "<Escape>", self._close_if_done, True
-        )  # escape-close only if process has completed
-        self.protocol("WM_DELETE_WINDOW", self._close)
-
-        self.update_idletasks()
-        self._start_listening()
-
-    def _start_listening(self):
-        def listen_stream(stream_name):
-            stream = getattr(self._proc, stream_name)
-            while True:
-                data = stream.readline()
-                self._event_queue.append((stream_name, data))
-                setattr(self, stream_name, getattr(self, stream_name) + data)
-                if data == "":
-                    break
-
-            self.returncode = self._proc.wait()
-
-        self._stdout_thread = threading.Thread(target=listen_stream, args=["stdout"], daemon=True)
-        self._stdout_thread.start()
-        if self._proc.stderr is not None:
-            self._stderr_thread = threading.Thread(
-                target=listen_stream, args=["stderr"], daemon=True
-            )
-            self._stderr_thread.start()
-
-        def poll_output_events():
-            if self._closed:
-                return
-
-            while len(self._event_queue) > 0:
-                stream_name, data = self._event_queue.popleft()
-                self.text.direct_insert("end", data, tags=(stream_name,))
-                self.text.see("end")
-
-            self.returncode = self._proc.poll()
-            if self.returncode == None:
-                self.after(200, poll_output_events)
-            else:
-                self.button["text"] = _("OK")
-                self.button.focus_set()
-                if self.returncode != 0:
-                    self.text.direct_insert(
-                        "end",
-                        "\n\nProcess failed, return code: %s\n" % self.returncode,
-                        ("stderr",),
-                    )
-                    self.update_idletasks()
-                    self.text.see("end")
-                elif self._autoclose:
-                    self._close()
-                else:
-                    self.text.direct_insert("end", "\n\n" + self._conclusion)
-                    self.update_idletasks()
-                    self.text.see("end")
-
-        poll_output_events()
-
-    def _close_if_done(self, event):
-        if self._proc.poll() is not None:
-            self._close(event)
-
-    def _close(self, event=None):
-        if self._proc.poll() is None:
-            if messagebox.askyesno(
-                _("Cancel the process?"),
-                _("The process is still running.\nAre you sure you want to cancel?"),
-                parent=None if running_on_mac_os() else self,
-            ):
-                # try gently first
-                try:
-                    if running_on_windows():
-                        os.kill(self._proc.pid, signal.CTRL_BREAK_EVENT)  # @UndefinedVariable
-                    else:
-                        os.kill(self._proc.pid, signal.SIGINT)
-
-                    self._proc.wait(2)
-                except subprocess.TimeoutExpired:
-                    if self._proc.poll() is None:
-                        # now let's be more concrete
-                        self._proc.kill()
-
-                self.cancelled = True
-                # Wait for threads to finish
-                self._stdout_thread.join(2)
-                if self._stderr_thread is not None:
-                    self._stderr_thread.join(2)
-
-                # fetch output about cancelling
-                while len(self._event_queue) > 0:
-                    stream_name, data = self._event_queue.popleft()
-                    self.text.direct_insert("end", data, tags=(stream_name,))
-                self.text.direct_insert("end", "\n\nPROCESS CANCELLED")
-                self.text.see("end")
-
-            else:
-                return
-        else:
-            self._closed = True
-            self.destroy()
-
-
 def get_busy_cursor():
     if running_on_windows():
         return "wait"
@@ -1898,9 +1852,11 @@ def open_path_in_system_file_manager(path):
 
 
 def _get_dialog_provider():
-    if platform.system() != "Linux":
+    if platform.system() != "Linux" or get_workbench().get_option("file.avoid_zenity"):
         return filedialog
 
+    import shutil
+
     if shutil.which("zenity"):
         return _ZenityDialogProvider
 
@@ -1910,42 +1866,57 @@ def _get_dialog_provider():
 
 def asksaveasfilename(**options):
     # https://tcl.tk/man/tcl8.6/TkCmd/getSaveFile.htm
-    _tweak_file_dialog_parent(options)
+    _check_dialog_parent(options)
     return _get_dialog_provider().asksaveasfilename(**options)
 
 
 def askopenfilename(**options):
     # https://tcl.tk/man/tcl8.6/TkCmd/getOpenFile.htm
-    _tweak_file_dialog_parent(options)
+    _check_dialog_parent(options)
     return _get_dialog_provider().askopenfilename(**options)
 
 
 def askopenfilenames(**options):
     # https://tcl.tk/man/tcl8.6/TkCmd/getOpenFile.htm
-    _tweak_file_dialog_parent(options)
+    _check_dialog_parent(options)
     return _get_dialog_provider().askopenfilenames(**options)
 
 
 def askdirectory(**options):
     # https://tcl.tk/man/tcl8.6/TkCmd/chooseDirectory.htm
-    _tweak_file_dialog_parent(options)
+    _check_dialog_parent(options)
     return _get_dialog_provider().askdirectory(**options)
 
 
-def _tweak_file_dialog_parent(options):
+def _check_dialog_parent(options):
+    if options.get("parent") and options.get("master"):
+        parent = options["parent"].winfo_toplevel()
+        master = options["master"].winfo_toplevel()
+        if parent is not master:
+            logger.warning(
+                "Dialog with different parent/master toplevels:\n%s",
+                "".join(traceback.format_stack()),
+            )
+    elif options.get("parent"):
+        parent = options["parent"].winfo_toplevel()
+        master = options["parent"].winfo_toplevel()
+    elif options.get("master"):
+        parent = options["master"].winfo_toplevel()
+        master = options["master"].winfo_toplevel()
+    else:
+        logger.warning("Dialog without parent:\n%s", "".join(traceback.format_stack()))
+        parent = tk._default_root
+        master = tk._default_root
+
+    options["parent"] = parent
+    options["master"] = master
+
     if running_on_mac_os():
         # used to require master/parent (https://bugs.python.org/issue34927)
         # but this is deprecated in Catalina (https://github.com/thonny/thonny/issues/840)
-        if "master" in options:
-            del options["master"]
-        if "parent" in options:
-            del options["parent"]
-    else:
-        if "parent" not in options:
-            if "master" in options:
-                options["parent"] = options["master"]
-            else:
-                options["parent"] = tk._default_root
+        # TODO: Consider removing this when upgrading from Tk 8.6.8
+        del options["master"]
+        del options["parent"]
 
 
 class _ZenityDialogProvider:
@@ -2024,6 +1995,8 @@ class _ZenityDialogProvider:
         if result.returncode == 0:
             return result.stdout.strip()
         else:
+            # TODO: log problems
+            print(result.stderr, file=sys.stderr)
             # could check stderr, but it may contain irrelevant warnings
             return None
 
@@ -2082,7 +2055,9 @@ def show_dialog(dlg, master=None, geometry=True, min_left=0, min_top=0):
         return
 
     if master is None:
-        master = tk._default_root
+        master = getattr(dlg, "parent", None) or getattr(dlg, "master", None) or tk._default_root
+
+    master = master.winfo_toplevel()
 
     get_workbench().event_generate("WindowFocusOut")
     # following order seems to give most smooth appearance
@@ -2097,14 +2072,14 @@ def show_dialog(dlg, master=None, geometry=True, min_left=0, min_top=0):
             assign_geometry(dlg, master, min_left, min_top)
         # dlg.wm_deiconify()
 
-    try:
-        dlg.grab_set()
-    except:
-        pass
-
     dlg.lift()
     dlg.focus_set()
 
+    try:
+        dlg.grab_set()
+    except TclError as e:
+        print("Can't grab:", e, file=sys.stderr)
+
     master.winfo_toplevel().wait_window(dlg)
     dlg.grab_release()
     master.winfo_toplevel().lift()
@@ -2181,13 +2156,13 @@ class MenuEx(tk.Menu):
                 else:
                     self.entryconfigure(i, state=tk.NORMAL)
 
-    def add(self, kind, cnf={}, **kw):
+    def add(self, itemType, cnf={}, **kw):
         cnf = cnf or kw
         tester = cnf.get("tester")
         if "tester" in cnf:
             del cnf["tester"]
 
-        super().add(kind, cnf)
+        super().add(itemType, cnf)
 
         itemdata = self.entryconfigure(self.index("end"))
         labeldata = itemdata.get("label")
@@ -2203,13 +2178,13 @@ class TextMenu(MenuEx):
         self.add_extra_items()
 
     def add_basic_items(self):
-        self.add_command(label=_("Cut"), command=self.on_cut, tester=self.can_cut)
-        self.add_command(label=_("Copy"), command=self.on_copy, tester=self.can_copy)
-        self.add_command(label=_("Paste"), command=self.on_paste, tester=self.can_paste)
+        self.add_command(label=tr("Cut"), command=self.on_cut, tester=self.can_cut)
+        self.add_command(label=tr("Copy"), command=self.on_copy, tester=self.can_copy)
+        self.add_command(label=tr("Paste"), command=self.on_paste, tester=self.can_paste)
 
     def add_extra_items(self):
         self.add_separator()
-        self.add_command(label=_("Select All"), command=self.on_select_all)
+        self.add_command(label=tr("Select All"), command=self.on_select_all)
 
     def on_cut(self):
         self.text.event_generate("<<Cut>>")
@@ -2246,13 +2221,18 @@ class TextMenu(MenuEx):
 
 
 def create_url_label(master, url, text=None):
+    import webbrowser
+
+    return create_action_label(master, text or url, lambda _: webbrowser.open(url))
+
+
+def create_action_label(master, text, click_handler, **kw):
     url_font = tkinter.font.nametofont("TkDefaultFont").copy()
     url_font.configure(underline=1)
     url_label = ttk.Label(
-        master, text=text if text else url, style="Url.TLabel", cursor="hand2", font=url_font
+        master, text=text, style="Url.TLabel", cursor="hand2", font=url_font, **kw
     )
-    url_label.grid()
-    url_label.bind("<Button-1>", lambda _: webbrowser.open(url))
+    url_label.bind("<Button-1>", click_handler)
     return url_label
 
 
@@ -2289,19 +2269,46 @@ def ems_to_pixels(x):
 _btn_padding = None
 
 
+def set_text_if_different(widget, text) -> bool:
+    if widget["text"] != text:
+        widget["text"] = text
+        return True
+    else:
+        return False
+
+
 def tr_btn(s):
     """Translates button caption, adds padding to make sure text fits"""
     global _btn_padding
     if _btn_padding is None:
         _btn_padding = get_button_padding()
 
-    return _btn_padding + _(s) + _btn_padding
+    return _btn_padding + tr(s) + _btn_padding
+
+
+def add_messagebox_parent_checker():
+    def wrap_with_parent_checker(original):
+        def wrapper(*args, **options):
+            _check_dialog_parent(options)
+            return original(*args, **options)
+
+        return wrapper
+
+    from tkinter import messagebox
+
+    for name in [
+        "showinfo",
+        "showwarning",
+        "showerror",
+        "askquestion",
+        "askokcancel",
+        "askyesno",
+        "askyesnocancel",
+        "askretrycancel",
+    ]:
+        fun = getattr(messagebox, name)
+        setattr(messagebox, name, wrap_with_parent_checker(fun))
 
 
 if __name__ == "__main__":
     root = tk.Tk()
-    closa = ClosableNotebook(root)
-    closa.add(ttk.Button(closa, text="B1"), text="B1")
-    closa.add(ttk.Button(closa, text="B2"), text="B2")
-    closa.grid()
-    root.mainloop()
