@@ -84,6 +84,7 @@ OBSOLETE_PLUGINS = [
     "thonnycontrib.circuitpython",
     "thonnycontrib.microbit",
     "thonnycontrib.esp",
+    "thonnycontrib.rpi_pico",
 ]
 
 
@@ -196,6 +197,7 @@ class Workbench(tk.Tk):
         self.bind_class("CodeViewText", "<<Modified>>", self.update_title, True)
         self.bind_class("CodeViewText", "<<TextChange>>", self.update_title, True)
         self.get_editor_notebook().bind("<<NotebookTabChanged>>", self.update_title, True)
+        self.get_editor_notebook().bind("<<NotebookTabChanged>>", self._update_toolbar, True)
         self.bind_all("<KeyPress>", self._on_all_key_presses, True)
         self.bind("<FocusOut>", self._on_focus_out, True)
         self.bind("<FocusIn>", self._on_focus_in, True)
@@ -549,8 +551,8 @@ class Workbench(tk.Tk):
                         client_socket.shutdown(socket.SHUT_WR)
                         raise PermissionError("Wrong secret")
 
-                except Exception:
-                    traceback.print_exc()
+                except Exception as e:
+                    logger.exception("Error in ipc server loop", exc_info=e)
 
         Thread(target=server_loop, daemon=True).start()
 
@@ -692,7 +694,7 @@ class Workbench(tk.Tk):
 
         margin = 10
         # Main frame functions as
-        # - a backgroud behind padding of main_pw, without this OS X leaves white border
+        # - a background behind padding of main_pw, without this OS X leaves white border
         # - a container to be hidden, when a view is maximized and restored when view is back home
         main_frame = ttk.Frame(self)  #
         self._main_frame = main_frame
@@ -1976,7 +1978,7 @@ class Workbench(tk.Tk):
     def get_toolbar_button(self, command_id):
         return self._toolbar_buttons[command_id]
 
-    def _update_toolbar(self) -> None:
+    def _update_toolbar(self, event=None) -> None:
         if self._destroyed or not hasattr(self, "_toolbar"):
             return
 
@@ -2174,7 +2176,10 @@ class Workbench(tk.Tk):
                 if tester:
                     try:
                         enabled = tester()
-                    except Exception:
+                    except Exception as e:
+                        logging.exception(
+                            "Could not check command tester for '%s'", item_data, exc_info=e
+                        )
                         traceback.print_exc()
                         enabled = False
 
@@ -2235,8 +2240,8 @@ class Workbench(tk.Tk):
                         if os.path.isfile(filename):
                             self.get_editor_notebook().show_file(filename)
 
-                except Exception:
-                    traceback.print_exc()
+                except Exception as e:
+                    logger.exception("Problem processing ipc request", exc_info=e)
 
             self.become_active_window()
         finally:
